#Copyright © 2018 Naturalpoint
#
#Licensed under the Apache License, Version 2.0 (the "License")
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

# OptiTrack NatNet direct depacketization library for Python 3.x

# modified by yuan-chu tai for mavproxy integration

import socket
import struct
import copy
import time
from MAVProxy.modules.mavproxy_optitrack import DataDescriptions
from MAVProxy.modules.mavproxy_optitrack import MoCapData

def trace( *args ):
    # uncomment the one you want to use
    #print( "".join(map(str,args)) )
    pass

#Used for Data Description functions
def trace_dd( *args ):
    # uncomment the one you want to use
    #print( "".join(map(str,args)) )
    pass

#Used for MoCap Frame Data functions
def trace_mf( *args ):
    # uncomment the one you want to use
    #print( "".join(map(str,args)) )
    pass

def get_message_id(data):
    message_id = int.from_bytes( data[0:2], byteorder='little' )
    return message_id


# Create structs for reading various object types to speed up parsing.
Vector2 = struct.Struct( '<ff' )
Vector3 = struct.Struct( '<fff' )
Quaternion = struct.Struct( '<ffff' )
FloatValue = struct.Struct( '<f' )
DoubleValue = struct.Struct( '<d' )
NNIntValue = struct.Struct( '<I')
FPCalMatrixRow = struct.Struct( '<ffffffffffff' )
FPCorners      = struct.Struct( '<ffffffffffff')

class NatNetClient:
    # print_level = 0 off
    # print_level = 1 on
    # print_level = >1 on / print every nth mocap frame
    print_level = 20
    
    def __init__( self ):
        # Change this value to the IP address of the NatNet server.
        self.server_ip_address = "127.0.0.1"

        # Change this value to the IP address of your local network interface
        self.local_ip_address = "127.0.0.1"

        # This should match the multicast address listed in Motive's streaming settings.
        self.multicast_address = "239.255.42.99"

        # NatNet Command channel
        self.command_port = 1510

        # NatNet Data channel
        self.data_port = 1511

        self.use_multicast = True

        # Set this to a callback method of your choice to receive per-rigid-body data at each frame.
        self.rigid_body_listener = None
        self.new_frame_listener  = None

        # Set Application Name
        self.__application_name = "Not Set"

        # NatNet stream version serveris capable of. This will be updated during initialization only.
        self.__nat_net_stream_version_server = [0,0,0,0]

        # NatNet stream version. This will be updated to the actual version the server is using during runtime.
        self.__nat_net_requested_version = [0,0,0,0]

        # server stream version. This will be updated to the actual version the server is using during initialization.
        self.__server_version = [0,0,0,0]

        # Lock values once run is called
        self.__is_locked = False

        # Server has the ability to change bitstream version
        self.__can_change_bitstream_version = False

        self.command_thread = None
        self.data_thread = None
        self.command_socket = None
        self.data_socket = None

        self.stop_threads=False


    # Client/server message ids
    NAT_CONNECT               = 0
    NAT_SERVERINFO            = 1
    NAT_REQUEST               = 2
    NAT_RESPONSE              = 3
    NAT_REQUEST_MODELDEF      = 4
    NAT_MODELDEF              = 5
    NAT_REQUEST_FRAMEOFDATA   = 6
    NAT_FRAMEOFDATA           = 7
    NAT_MESSAGESTRING         = 8
    NAT_DISCONNECT            = 9
    NAT_KEEPALIVE             = 10
    NAT_UNRECOGNIZED_REQUEST  = 100
    NAT_UNDEFINED             = 999999.9999


    def set_client_address(self, local_ip_address):
        if not self.__is_locked:
            self.local_ip_address = local_ip_address

    def get_client_address(self):
        return self.local_ip_address

    def set_server_address(self,server_ip_address):
        if not self.__is_locked:
            self.server_ip_address = server_ip_address

    def get_server_address(self):
        return self.server_ip_address


    def set_use_multicast(self, use_multicast):
        if not self.__is_locked:
            self.use_multicast = use_multicast

    def can_change_bitstream_version(self):
        return self.__can_change_bitstream_version

    def set_nat_net_version(self, major, minor):
        """checks to see if stream version can change, then changes it with position reset"""
        return_code = -1
        if self.__can_change_bitstream_version and \
            (major != self.__nat_net_requested_version[0]) and\
            (minor != self.__nat_net_requested_version[1]):
            sz_command = "Bitstream,%1.1d.%1.1d"%(major, minor)
            return_code = self.send_command(sz_command)
            if return_code >=0:
                self.__nat_net_requested_version[0] = major
                self.__nat_net_requested_version[1] = minor
                self.__nat_net_requested_version[2] = 0
                self.__nat_net_requested_version[3] = 0
                print("changing bitstream MAIN")
                # get original output state
                #print_results = self.get_print_results()
                #turn off output
                #self.set_print_results(False)
                # force frame send and play reset
                self.send_command("TimelinePlay")
                time.sleep(0.1)
                tmpCommands=["TimelinePlay",
                    "TimelineStop",
                    "SetPlaybackCurrentFrame,0",
                    "TimelineStop"]
                self.send_commands(tmpCommands,False)
                time.sleep(2)
                #reset to original output state
                #self.set_print_results(print_results)
        return return_code


    def get_major(self):
        return self.__nat_net_requested_version[0]

    def get_minor(self):
        return self.__nat_net_requested_version[1]

    def set_print_level(self, print_level=0):
        if(print_level >=0):
            self.print_level = print_level
        return self.print_level

    def get_print_level(self):
        return self.print_level


    def connected(self):
        ret_value = True
        # check sockets
        if self.command_socket == None:
            ret_value = False
        elif self.data_socket ==None:
            ret_value = False
        # check versions
        elif self.get_application_name() == "Not Set":
            ret_value = False
        elif (self.__server_version[0] == 0) and\
            (self.__server_version[1] == 0) and\
            (self.__server_version[2] == 0) and\
            (self.__server_version[3] == 0):
            ret_value = False
        return ret_value


    # Create a command socket to attach to the NatNet stream
    def __create_command_socket( self):
        result = None
        if self.use_multicast :
            # Multicast case
            result = socket.socket( socket.AF_INET, socket.SOCK_DGRAM, 0 )
            # allow multiple clients on same machine to use multicast group address/port
            result.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                result.bind( ('', 0) )
            except socket.error as msg:
                print("ERROR: command socket error occurred:\n%s" %msg)
                print("Check Motive/Server mode requested mode agreement.  You requested Multicast ")
                result = None
            except  socket.herror:
                print("ERROR: command socket herror occurred")
                result = None
            except  socket.gaierror:
                print("ERROR: command socket gaierror occurred")
                result = None
            except  socket.timeout:
                print("ERROR: command socket timeout occurred. Server not responding")
                result = None
            # set to broadcast mode
            result.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            # set timeout to allow for keep alive messages
            result.settimeout(2.0)
        else:
            # Unicast case
            result = socket.socket( socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            try:
                result.bind( (self.local_ip_address, 0) )
            except socket.error as msg:
                print("ERROR: command socket error occurred:\n%s" %msg)
                print("Check Motive/Server mode requested mode agreement.  You requested Unicast ")
                result = None
            except socket.herror:
                print("ERROR: command socket herror occurred")
                result = None
            except socket.gaierror:
                print("ERROR: command socket gaierror occurred")
                result = None
            except socket.timeout:
                print("ERROR: command socket timeout occurred. Server not responding")
                result = None

            # set timeout to allow for keep alive messages
            result.settimeout(2.0)
            result.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        return result

    # Create a data socket to attach to the NatNet stream
    def __create_data_socket( self, port):
        result = None

        if self.use_multicast:
            # Multicast case
            result = socket.socket( socket.AF_INET,     # Internet
                                  socket.SOCK_DGRAM,
                                  0)    # UDP
            result.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            result.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(self.multicast_address) + socket.inet_aton(self.local_ip_address))
            try:
                result.bind( (self.local_ip_address, port) )
            except socket.error as msg:
                print("ERROR: data socket error occurred:\n%s" %msg)
                print("  Check Motive/Server mode requested mode agreement.  You requested Multicast ")
                result = None
            except socket.herror:
                print("ERROR: data socket herror occurred")
                result = None
            except socket.gaierror:
                print("ERROR: data socket gaierror occurred")
                result = None
            except socket.timeout:
                print("ERROR: data socket timeout occurred. Server not responding")
                result = None
        else:
            # Unicast case
            result = socket.socket( socket.AF_INET,     # Internet
                                  socket.SOCK_DGRAM,
                                  socket.IPPROTO_UDP)
            result.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            #result.bind( (self.local_ip_address, port) )
            try:
                result.bind( ('', 0) )
            except socket.error as msg:
                print("ERROR: data socket error occurred:\n%s" %msg)
                print("Check Motive/Server mode requested mode agreement.  You requested Unicast ")
                result = None
            except socket.herror:
                print("ERROR: data socket herror occurred")
                result = None
            except socket.gaierror:
                print("ERROR: data socket gaierror occurred")
                result = None
            except socket.timeout:
                print("ERROR: data socket timeout occurred. Server not responding")
                result = None
            
            if(self.multicast_address != "255.255.255.255"):
                result.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(self.multicast_address) + socket.inet_aton(self.local_ip_address))

        return result

    # Unpack a rigid body object from a data packet
    def __unpack_rigid_body( self, data, major, minor, rb_num):
        offset = 0

        # ID (4 bytes)
        new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4

        trace_mf( "RB: %3.1d ID: %3.1d"% (rb_num, new_id))

        # Position and orientation
        pos = Vector3.unpack( data[offset:offset+12] )
        offset += 12
        trace_mf( "\tPosition    : [%3.2f, %3.2f, %3.2f]"% (pos[0], pos[1], pos[2] ))

        rot = Quaternion.unpack( data[offset:offset+16] )
        offset += 16
        trace_mf( "\tOrientation : [%3.2f, %3.2f, %3.2f, %3.2f]"% (rot[0], rot[1], rot[2], rot[3] ))

        rigid_body = MoCapData.RigidBody(new_id, pos, rot)

        # RB Marker Data ( Before version 3.0.  After Version 3.0 Marker data is in description )
        if( major < 3  and major != 0) :
            # Marker count (4 bytes)
            marker_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            marker_count_range = range( 0, marker_count )
            trace_mf( "\tMarker Count:", marker_count )

            rb_marker_list=[]
            for i in marker_count_range:
                rb_marker_list.append(MoCapData.RigidBodyMarker())

            # Marker positions
            for i in marker_count_range:
                pos = Vector3.unpack( data[offset:offset+12] )
                offset += 12
                trace_mf( "\tMarker", i, ":", pos[0],",", pos[1],",", pos[2] )
                rb_marker_list[i].pos=pos


            if major >= 2:
                # Marker ID's
                for i in marker_count_range:
                    marker_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
                    offset += 4
                    trace_mf( "\tMarker ID", i, ":", marker_id )
                    rb_marker_list[i].id=marker_id

                # Marker sizes
                for i in marker_count_range:
                    size = FloatValue.unpack( data[offset:offset+4] )
                    offset += 4
                    trace_mf( "\tMarker Size", i, ":", size[0] )
                    rb_marker_list[i].size=size

            for i in marker_count_range:
                rigid_body.add_rigid_body_marker(rb_marker_list[i])
        if major >= 2 :
            marker_error, = FloatValue.unpack( data[offset:offset+4] )
            offset += 4
            trace_mf( "\tMarker Error: %3.2f"% marker_error )
            rigid_body.error = marker_error

        tracking_valid = True
        # Version 2.6 and later
        if ( ( major == 2 ) and ( minor >= 6 ) ) or major > 2 :
            param, = struct.unpack( 'h', data[offset:offset+2] )
            tracking_valid = ( param & 0x01 ) != 0
            offset += 2
            is_valid_str='False'
            if tracking_valid:
                is_valid_str = 'True'
            trace_mf( "\tTracking Valid: %s"%is_valid_str)
            if tracking_valid:
                rigid_body.tracking_valid = True
            else:
                rigid_body.tracking_valid = False

        # Send information to any listener.
        if self.rigid_body_listener is not None:
            self.rigid_body_listener( new_id, pos, rot, tracking_valid )

        return offset, rigid_body

    # Unpack a skeleton object from a data packet
    def __unpack_skeleton( self, data, major, minor):

        offset = 0
        new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "ID:", new_id )
        skeleton = MoCapData.Skeleton(new_id)

        rigid_body_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "Rigid Body Count : %3.1d"% rigid_body_count )
        for rb_num in range( 0, rigid_body_count ):
            offset_tmp, rigid_body = self.__unpack_rigid_body( data[offset:], major, minor, rb_num )
            skeleton.add_rigid_body(rigid_body)
            offset+=offset_tmp

        return offset, skeleton

    #Unpack Mocap Data Functions
    def __unpack_frame_prefix_data( self, data):
        offset = 0
        # Frame number (4 bytes)
        frame_number = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "Frame #:", frame_number )
        frame_prefix_data=MoCapData.FramePrefixData(frame_number)
        return offset, frame_prefix_data

    def __unpack_marker_set_data( self, data, packet_size, major, minor):
        marker_set_data=MoCapData.MarkerSetData()
        offset = 0
        # Marker set count (4 bytes)
        marker_set_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "Marker Set Count:", marker_set_count )

        for i in range( 0, marker_set_count ):
            marker_data = MoCapData.MarkerData()
            # Model name
            model_name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( model_name ) + 1
            trace_mf( "Model Name      : ", model_name.decode( 'utf-8' ) )
            marker_data.set_model_name(model_name)
            # Marker count (4 bytes)
            marker_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_mf( "Marker Count    : ", marker_count )

            for j in range( 0, marker_count ):
                pos = Vector3.unpack( data[offset:offset+12] )
                offset += 12
                trace_mf( "\tMarker %3.1d : [%3.2f,%3.2f,%3.2f]"%( j, pos[0], pos[1], pos[2] ))
                marker_data.add_pos(pos)
            marker_set_data.add_marker_data(marker_data)

        # Unlabeled markers count (4 bytes)
        unlabeled_markers_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "Unlabeled Markers Count:", unlabeled_markers_count )

        for i in range( 0, unlabeled_markers_count ):
            pos = Vector3.unpack( data[offset:offset+12] )
            offset += 12
            trace_mf( "\tMarker %3.1d : [%3.2f,%3.2f,%3.2f]"%( i, pos[0], pos[1], pos[2] ))
            marker_set_data.add_unlabeled_marker(pos)
        return offset, marker_set_data

    def __unpack_rigid_body_data( self, data, packet_size, major, minor):
        rigid_body_data = MoCapData.RigidBodyData()
        offset = 0
        # Rigid body count (4 bytes)
        rigid_body_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_mf( "Rigid Body Count:", rigid_body_count )

        for i in range( 0, rigid_body_count ):
            offset_tmp, rigid_body = self.__unpack_rigid_body( data[offset:], major, minor, i )
            offset += offset_tmp
            rigid_body_data.add_rigid_body(rigid_body)

        return offset, rigid_body_data


    def __unpack_skeleton_data( self, data, packet_size, major, minor):
        skeleton_data = MoCapData.SkeletonData()

        offset = 0
        # Version 2.1 and later
        skeleton_count = 0
        if( ( major == 2 and minor > 0 ) or major > 2 ):
            skeleton_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_mf( "Skeleton Count:", skeleton_count )
            for _ in range( 0, skeleton_count ):
                rel_offset, skeleton = self.__unpack_skeleton( data[offset:], major, minor )
                offset += rel_offset
                skeleton_data.add_skeleton(skeleton)

        return offset, skeleton_data

    def __decode_marker_id(self, new_id):
        model_id = 0
        marker_id = 0
        model_id = new_id >> 16
        marker_id = new_id & 0x0000ffff
        return model_id, marker_id

    def __unpack_labeled_marker_data( self, data, packet_size, major, minor):
        labeled_marker_data = MoCapData.LabeledMarkerData()
        offset = 0
        # Labeled markers (Version 2.3 and later)
        labeled_marker_count = 0
        if( ( major == 2 and minor > 3 ) or major > 2 ):
            labeled_marker_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_mf( "Labeled Marker Count:", labeled_marker_count )
            for _ in range( 0, labeled_marker_count ):
                model_id = 0
                marker_id = 0
                tmp_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4
                model_id, marker_id = self.__decode_marker_id(tmp_id)
                pos = Vector3.unpack( data[offset:offset+12] )
                offset += 12
                size = FloatValue.unpack( data[offset:offset+4] )
                offset += 4
                trace_mf("ID     : [MarkerID: %3.1d] [ModelID: %3.1d]"%(marker_id,model_id))
                trace_mf("  pos  : [%3.2f, %3.2f, %3.2f]"%(pos[0],pos[1],pos[2]))
                trace_mf("  size : [%3.2f]"%size)


                # Version 2.6 and later
                param = 0
                if( ( major == 2 and minor >= 6 ) or major > 2):
                    param, = struct.unpack( 'h', data[offset:offset+2] )
                    offset += 2
                    #occluded = ( param & 0x01 ) != 0
                    #point_cloud_solved = ( param & 0x02 ) != 0
                    #model_solved = ( param & 0x04 ) != 0

                # Version 3.0 and later
                residual = 0.0
                if major >= 3 :
                    residual, = FloatValue.unpack( data[offset:offset+4] )
                    offset += 4
                    trace_mf( "  err  : [%3.2f]"% residual )

                labeled_marker = MoCapData.LabeledMarker(tmp_id,pos,size,param, residual)
                labeled_marker_data.add_labeled_marker(labeled_marker)

        return offset, labeled_marker_data

    def __unpack_force_plate_data( self, data, packet_size, major, minor):
        force_plate_data = MoCapData.ForcePlateData()
        n_frames_show_max = 4
        offset = 0
        # Force Plate data (version 2.9 and later)
        force_plate_count = 0
        if( ( major == 2 and minor >= 9 ) or major > 2 ):
            force_plate_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_mf( "Force Plate Count:", force_plate_count )
            for i in range( 0, force_plate_count ):
                # ID
                force_plate_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4
                force_plate = MoCapData.ForcePlate(force_plate_id)

                # Channel Count
                force_plate_channel_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4

                trace_mf( "\tForce Plate %3.1d ID: %3.1d Num Channels: %3.1d"% (i, force_plate_id, force_plate_channel_count ))

                # Channel Data
                for j in range( force_plate_channel_count ):
                    fp_channel_data = MoCapData.ForcePlateChannelData()
                    force_plate_channel_frame_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
                    offset += 4
                    out_string="\tChannel %3.1d: "%( j )
                    out_string+="  %3.1d Frames - Frame Data: "%(force_plate_channel_frame_count)

                    # Force plate frames
                    n_frames_show = min(force_plate_channel_frame_count, n_frames_show_max)
                    for k in range( force_plate_channel_frame_count ):
                        force_plate_channel_val = FloatValue.unpack( data[offset:offset+4] )
                        offset += 4
                        fp_channel_data.add_frame_entry(force_plate_channel_val)

                        if k < n_frames_show:
                            out_string += "%3.2f "%(force_plate_channel_val)
                    if n_frames_show < force_plate_channel_frame_count:
                        out_string += " showing %3.1d of %3.1d frames"%(n_frames_show, force_plate_channel_frame_count)
                    trace_mf( "%s"% out_string )
                    force_plate.add_channel_data(fp_channel_data)
                force_plate_data.add_force_plate(force_plate)
        return offset, force_plate_data

    def __unpack_device_data( self, data, packet_size, major, minor):
        device_data = MoCapData.DeviceData()
        n_frames_show_max = 4
        offset = 0
        # Device data (version 2.11 and later)
        device_count = 0
        if ( major == 2 and minor >= 11 ) or (major > 2) :
            device_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_mf( "Device Count:", device_count )
            for i in range( 0, device_count ):

                # ID
                device_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4
                device = MoCapData.Device(device_id)
                # Channel Count
                device_channel_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4

                trace_mf( "\tDevice %3.1d      ID: %3.1d Num Channels: %3.1d"% (i, device_id, device_channel_count ))

                # Channel Data
                for j in range( 0, device_channel_count ):
                    device_channel_data = MoCapData.DeviceChannelData()
                    device_channel_frame_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
                    offset += 4
                    out_string="\tChannel %3.1d "% (j)
                    out_string+="  %3.1d Frames - Frame Data: "%(device_channel_frame_count)

                    # Device Frame Data
                    n_frames_show = min(device_channel_frame_count, n_frames_show_max)
                    for k in range( 0, device_channel_frame_count ):
                        device_channel_val = int.from_bytes( data[offset:offset+4], byteorder='little' )
                        device_channel_val = FloatValue.unpack( data[offset:offset+4] )
                        offset += 4
                        if k < n_frames_show:
                            out_string += "%3.2f "%(device_channel_val)

                        device_channel_data.add_frame_entry(device_channel_val)
                    if n_frames_show < device_channel_frame_count:
                        out_string += " showing %3.1d of %3.1d frames"%(n_frames_show, device_channel_frame_count)
                    trace_mf( "%s"% out_string )
                    device.add_channel_data(device_channel_data)
                device_data.add_device(device)
        return offset, device_data

    def __unpack_frame_suffix_data( self, data, packet_size, major, minor):
        frame_suffix_data = MoCapData.FrameSuffixData()
        offset = 0

        # Timecode
        timecode = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        frame_suffix_data.timecode = timecode

        timecode_sub = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        frame_suffix_data.timecode_sub = timecode_sub

        # Timestamp (increased to double precision in 2.7 and later)
        if ( major == 2 and minor >= 7 ) or (major > 2 ):
            timestamp, = DoubleValue.unpack( data[offset:offset+8] )
            offset += 8
        else:
            timestamp, = FloatValue.unpack( data[offset:offset+4] )
            offset += 4
        trace_mf("Timestamp : %3.2f"%timestamp)
        frame_suffix_data.timestamp = timestamp

        # Hires Timestamp (Version 3.0 and later)
        if major >= 3 :
            stamp_camera_mid_exposure = int.from_bytes( data[offset:offset+8], byteorder='little' )
            trace_mf("Mid-exposure timestamp         : %3.1d"%stamp_camera_mid_exposure)
            offset += 8
            frame_suffix_data.stamp_camera_mid_exposure = stamp_camera_mid_exposure

            stamp_data_received = int.from_bytes( data[offset:offset+8], byteorder='little' )
            offset += 8
            frame_suffix_data.stamp_data_received = stamp_data_received
            trace_mf("Camera data received timestamp : %3.1d"%stamp_data_received)

            stamp_transmit = int.from_bytes( data[offset:offset+8], byteorder='little' )
            offset += 8
            trace_mf("Transmit timestamp             : %3.1d"%stamp_transmit)
            frame_suffix_data.stamp_transmit = stamp_transmit


        # Frame parameters
        param, = struct.unpack( 'h', data[offset:offset+2] )
        is_recording = ( param & 0x01 ) != 0
        tracked_models_changed = ( param & 0x02 ) != 0
        offset += 2
        frame_suffix_data.param = param
        frame_suffix_data.is_recording = is_recording
        frame_suffix_data.tracked_models_changed = tracked_models_changed

        return offset, frame_suffix_data


    # Unpack data from a motion capture frame message
    def __unpack_mocap_data( self, data : bytes, packet_size, major, minor):
        mocap_data = MoCapData.MoCapData()
        trace_mf( "MoCap Frame Begin\n-----------------" )
        data = memoryview( data )
        offset = 0
        rel_offset = 0

        #Frame Prefix Data
        rel_offset, frame_prefix_data = self.__unpack_frame_prefix_data(data[offset:])
        offset += rel_offset
        mocap_data.set_prefix_data(frame_prefix_data)
        frame_number = frame_prefix_data.frame_number

        #Marker Set Data
        rel_offset, marker_set_data =self.__unpack_marker_set_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_marker_set_data(marker_set_data)
        marker_set_count = marker_set_data.get_marker_set_count()
        unlabeled_markers_count = marker_set_data.get_unlabeled_marker_count()

        # Rigid Body Data
        rel_offset, rigid_body_data = self.__unpack_rigid_body_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_rigid_body_data(rigid_body_data)
        rigid_body_count = rigid_body_data.get_rigid_body_count()

        # Skeleton Data
        rel_offset, skeleton_data = self.__unpack_skeleton_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_skeleton_data(skeleton_data)
        skeleton_count = skeleton_data.get_skeleton_count()

        # Labeled Marker Data
        rel_offset, labeled_marker_data = self.__unpack_labeled_marker_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_labeled_marker_data(labeled_marker_data)
        labeled_marker_count = labeled_marker_data.get_labeled_marker_count()

        # Force Plate Data
        rel_offset, force_plate_data = self.__unpack_force_plate_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_force_plate_data(force_plate_data)

        # Device Data
        rel_offset,device_data = self.__unpack_device_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_device_data(device_data)

        # Frame Suffix Data
        #rel_offset, timecode, timecode_sub, timestamp, is_recording, tracked_models_changed = \
        rel_offset, frame_suffix_data =   self.__unpack_frame_suffix_data(data[offset:], (packet_size - offset),major, minor)
        offset += rel_offset
        mocap_data.set_suffix_data(frame_suffix_data)


        timecode = frame_suffix_data.timecode
        timecode_sub= frame_suffix_data.timecode_sub
        timestamp = frame_suffix_data.timestamp
        is_recording = frame_suffix_data.is_recording
        tracked_models_changed = frame_suffix_data.tracked_models_changed
        # Send information to any listener.
        if self.new_frame_listener is not None:
            data_dict={}
            data_dict["frame_number"]=frame_number
            data_dict[ "marker_set_count"] = marker_set_count
            data_dict[ "unlabeled_markers_count"] = unlabeled_markers_count
            data_dict[ "rigid_body_count"] = rigid_body_count
            data_dict[ "skeleton_count"] =skeleton_count
            data_dict[ "labeled_marker_count"] = labeled_marker_count
            data_dict[ "timecode"] = timecode
            data_dict[ "timecode_sub"] = timecode_sub
            data_dict[ "timestamp"] = timestamp
            data_dict[ "is_recording"] = is_recording
            data_dict[ "tracked_models_changed"] = tracked_models_changed

            self.new_frame_listener( data_dict )
        trace_mf( "MoCap Frame End\n-----------------" )
        return offset, mocap_data


    # Unpack a marker set description packet
    def __unpack_marker_set_description( self, data, major, minor):
        ms_desc = DataDescriptions.MarkerSetDescription()

        offset = 0

        name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
        offset += len( name ) + 1
        trace_dd( "Marker Set Name: %s" % (name.decode( 'utf-8' )) )
        ms_desc.set_name(name)

        marker_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_dd( "Marker Count : %3.1d" % marker_count)
        for i in range( 0, marker_count ):
            name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( name ) + 1
            trace_dd( "\t%2.1d Marker Name: %s"%(i, name.decode( 'utf-8' ) ))
            ms_desc.add_marker_name(name)

        return offset, ms_desc

    # Unpack a rigid body description packet
    def __unpack_rigid_body_description( self, data, major, minor):
        rb_desc=DataDescriptions.RigidBodyDescription()
        offset = 0

        # Version 2.0 or higher
        if (major >= 2) or (major == 0):
            name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( name ) + 1
            rb_desc.set_name(name)
            trace_dd( "\tRigid Body Name   : ", name.decode( 'utf-8' ) )

        # ID
        new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        rb_desc.set_id(new_id)
        trace_dd( "\tID                : ", str(new_id))

        #Parent ID
        parent_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        rb_desc.set_parent_id(parent_id)
        trace_dd( "\tParent ID         : ", parent_id)

        # Position Offsets
        pos = Vector3.unpack( data[offset:offset+12] )
        offset += 12
        rb_desc.set_pos(pos[0],pos[1],pos[2])

        trace_dd( "\tPosition          : [%3.2f, %3.2f, %3.2f]"% (pos[0], pos[1], pos[2] ))

        # Version 3.0 and higher, rigid body marker information contained in description
        if (major >= 3) or (major == 0) :
            # Marker Count
            marker_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_dd( "\tNumber of Markers : ", marker_count )

            marker_count_range = range( 0, marker_count )
            offset1 = offset
            offset2 = offset1 + (12*marker_count)
            offset3 = offset2 + (4*marker_count)
            # Marker Offsets X,Y,Z
            marker_name=""
            for marker in marker_count_range:
                # Offset
                marker_offset = Vector3.unpack(data[offset1:offset1+12])
                offset1 +=12

                # Active Label
                active_label = int.from_bytes(data[offset2:offset2+4],byteorder = 'little')
                offset2 += 4

                #Marker Name
                if (major >= 4) or (major == 0):
                    # markername
                    marker_name, separator, remainder = bytes(data[offset3:]).partition( b'\0' )
                    marker_name = marker_name.decode( 'utf-8' )
                    offset3 += len( marker_name ) + 1

                rb_marker=DataDescriptions.RBMarker(marker_name,active_label,marker_offset)
                rb_desc.add_rb_marker(rb_marker)
                trace_dd( "\t%3.1d Marker Label: %s Position: [%3.2f %3.2f %3.2f] %s" % (marker,active_label,\
                   marker_offset[0], marker_offset[1], marker_offset[2],marker_name ))

            offset = offset3
        
        trace_dd("\tunpack_rigid_body_description processed bytes: ", offset)
        return offset, rb_desc

    # Unpack a skeleton description packet
    def __unpack_skeleton_description( self, data, major, minor):
        skeleton_desc = DataDescriptions.SkeletonDescription()
        offset = 0

        #Name
        name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
        offset += len( name ) + 1
        skeleton_desc.set_name(name)
        trace_dd( "Name : %s"% name.decode( 'utf-8' ) )

        #ID
        new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        skeleton_desc.set_id(new_id)
        trace_dd( "ID : %3.1d"% new_id )

        # # of RigidBodies
        rigid_body_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_dd( "Rigid Body (Bone) Count : %3.1d" % rigid_body_count)

        # Loop over all Rigid Bodies
        for i in range( 0, rigid_body_count ):
            trace_dd("Rigid Body (Bone) ", i)
            offset_tmp, rb_desc_tmp = self.__unpack_rigid_body_description( data[offset:], major, minor )
            offset+= offset_tmp
            skeleton_desc.add_rigid_body_description(rb_desc_tmp)
        return offset, skeleton_desc

    def __unpack_force_plate_description(self, data, major, minor):
        fp_desc = None
        offset = 0
        if major >= 3:
            fp_desc = DataDescriptions.ForcePlateDescription()
            # ID
            new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            fp_desc.set_id(new_id)
            trace_dd("\tID : ", str(new_id))

            # Serial Number
            serial_number, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( serial_number ) + 1
            fp_desc.set_serial_number(serial_number)
            trace_dd( "\tSerial Number : ", serial_number.decode( 'utf-8' ) )

            # Dimensions
            f_width = FloatValue.unpack( data[offset:offset+4])
            offset += 4
            trace_dd( "\tWidth  : %3.2f"% f_width)
            f_length = FloatValue.unpack( data[offset:offset+4])
            offset += 4
            fp_desc.set_dimensions(f_width[0], f_length[0])
            trace_dd( "\tLength : %3.2f"% f_length)

            # Origin
            origin = Vector3.unpack( data[offset:offset+12] )
            offset += 12
            fp_desc.set_origin(origin[0],origin[1],origin[2])
            trace_dd( "\tOrigin : %3.2f, %3.2f, %3.2f"%( origin[0], origin[1], origin[2] ))

            # Calibration Matrix 12x12 floats
            trace_dd("Cal Matrix:")
            cal_matrix_tmp= [[0.0 for col in range(12)] for row in range(12)]

            for i in range(0,12):
                cal_matrix_row=FPCalMatrixRow.unpack(data[offset:offset+(12*4)])
                trace_dd("\t%3.1d %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e %3.3e" % (i
                      , cal_matrix_row[0], cal_matrix_row[1], cal_matrix_row[2], cal_matrix_row[3]
                      , cal_matrix_row[4], cal_matrix_row[5], cal_matrix_row[6], cal_matrix_row[7]
                      , cal_matrix_row[8], cal_matrix_row[9], cal_matrix_row[10], cal_matrix_row[11]))
                cal_matrix_tmp[i] = copy.deepcopy(cal_matrix_row)
                offset += (12*4)
            fp_desc.set_cal_matrix(cal_matrix_tmp)
            # Corners 4x3 floats
            corners = FPCorners.unpack(data[offset:offset + (12*4)])
            offset += (12*4)
            o_2=0
            trace_dd("Corners:")
            corners_tmp = [[0.0 for col in range(3)] for row in range(4)]
            for i in range(0,4):
                trace_dd("\t%3.1d %3.3e %3.3e %3.3e"%(i, corners[o_2], corners[o_2+1], corners[o_2+2]))
                corners_tmp[i][0]=corners[o_2]
                corners_tmp[i][1]=corners[o_2+1]
                corners_tmp[i][2]=corners[o_2+2]
                o_2+=3
            fp_desc.set_corners(corners_tmp)

            # Plate Type int
            plate_type = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            fp_desc.set_plate_type(plate_type)
            trace_dd ("Plate Type : ", plate_type)

            # Channel Data Type int
            channel_data_type = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            fp_desc.set_channel_data_type(channel_data_type)
            trace_dd("Channel Data Type : ", channel_data_type)

            # Number of Channels int
            num_channels = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            trace_dd("Number of Channels : ", num_channels)

            # Channel Names list of NoC strings
            for i in range(0, num_channels):
                channel_name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
                offset += len( channel_name ) + 1
                trace_dd( "\tChannel Name %3.1d: %s"%(i, channel_name.decode( 'utf-8' ) ))
                fp_desc.add_channel_name(channel_name)

        trace_dd("unpackForcePlate processed ", offset, " bytes")
        return offset, fp_desc

    def __unpack_device_description(self, data, major, minor):
        device_desc=None
        offset = 0
        if major >= 3:
            # new_id
            new_id = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            trace_dd("\tID : ", str(new_id))

            # Name
            name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( name ) + 1
            trace_dd( "\tName : ", name.decode( 'utf-8' ) )

            # Serial Number
            serial_number, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( serial_number ) + 1
            trace_dd( "\tSerial Number : ", serial_number.decode( 'utf-8' ) )


            # Device Type int
            device_type = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            trace_dd ("Device Type : ", device_type)

            # Channel Data Type int
            channel_data_type = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            trace_dd("Channel Data Type : ", channel_data_type)

            device_desc = DataDescriptions.DeviceDescription(new_id,name,serial_number,device_type,channel_data_type)

            # Number of Channels int
            num_channels = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset+=4
            trace_dd("Number of Channels ", num_channels)

            # Channel Names list of NoC strings
            for i in range(0, num_channels):
                channel_name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
                offset += len( channel_name ) + 1
                device_desc.add_channel_name(channel_name)
                trace_dd( "\tChannel ",i," Name : ", channel_name.decode( 'utf-8' ) )

        trace_dd("unpack_device_description processed ", offset, " bytes")
        return offset, device_desc

    def __unpack_camera_description(self, data, major, minor):
        offset = 0
        # Name
        name, separator, remainder = bytes(data[offset:]).partition( b'\0' )
        offset += len( name ) + 1
        trace_dd( "\tName       : %s"% name.decode( 'utf-8' ) )
        # Position
        position = Vector3.unpack( data[offset:offset+12] )
        offset += 12
        trace_dd( "\tPosition   : [%3.2f, %3.2f, %3.2f]"% (position[0], position[1], position[2] ))

        # Orientation
        orientation = Quaternion.unpack( data[offset:offset+16] )
        offset += 16
        trace_dd( "\tOrientation: [%3.2f, %3.2f, %3.2f, %3.2f]"% (orientation[0], orientation[1], orientation[2], orientation[3] ))
        trace_dd("unpack_camera_description processed %3.1d bytes"% offset)

        camera_desc=DataDescriptions.CameraDescription(name, position, orientation)
        return offset, camera_desc


    # Unpack a data description packet
    def __unpack_data_descriptions( self, data : bytes, packet_size, major, minor):
        data_descs = DataDescriptions.DataDescriptions()
        offset = 0
        # # of data sets to process
        dataset_count = int.from_bytes( data[offset:offset+4], byteorder='little' )
        offset += 4
        trace_dd("Dataset Count : ", str(dataset_count))
        for i in range( 0, dataset_count ):
            trace_dd("Dataset ", str(i))
            data_type = int.from_bytes( data[offset:offset+4], byteorder='little' )
            offset += 4
            data_tmp=None
            if data_type == 0 :
                trace_dd("Type: 0 Markerset")
                offset_tmp, data_tmp = self.__unpack_marker_set_description( data[offset:], major, minor )
            elif data_type == 1 :
                trace_dd("Type: 1 Rigid Body")
                offset_tmp, data_tmp = self.__unpack_rigid_body_description( data[offset:], major, minor )
            elif data_type == 2 :
                trace_dd("Type: 2 Skeleton")
                offset_tmp, data_tmp = self.__unpack_skeleton_description( data[offset:], major, minor )
            elif data_type == 3 :
                trace_dd("Type: 3 Force Plate")
                offset_tmp, data_tmp = self.__unpack_force_plate_description(data[offset:], major, minor)
            elif data_type == 4 :
                trace_dd("Type: 4 Device")
                offset_tmp, data_tmp = self.__unpack_device_description(data[offset:], major, minor)
            elif data_type == 5 :
                trace_dd("Type: 5 Camera")
                offset_tmp, data_tmp = self.__unpack_camera_description(data[offset:], major, minor)
            else:
                print("Type: " + str(data_type) + " UNKNOWN")
                print("ERROR: Type decode failure" )
                print("\t"+ str(i + 1) +" datasets processed of " + str(dataset_count))
                print("\t "+ str(offset) +" bytes processed of " + str(packet_size) )
                print("\tPACKET DECODE STOPPED")
                return offset
            offset += offset_tmp
            data_descs.add_data(data_tmp)
            trace_dd("\t"+ str(i) +" datasets processed of " + str(dataset_count))
            trace_dd("\t "+ str(offset) +" bytes processed of " + str(packet_size) )

        return offset, data_descs

    # __unpack_server_info is for local use of the client
    # and will update the values for the versions/ NatNet capabilities
    # of the server.
    def __unpack_server_info(self, data, packet_size, major, minor):
        offset = 0
        # Server name
        #szName = data[offset: offset+256]
        self.__application_name, separator, remainder = bytes(data[offset: offset+256]).partition( b'\0' )
        self.__application_name=str(self.__application_name, "utf-8")
        offset += 256
        # Server Version info
        server_version = struct.unpack( 'BBBB', data[offset:offset+4] )
        offset += 4
        self.__server_version[0] = server_version[0]
        self.__server_version[1] = server_version[1]
        self.__server_version[2] = server_version[2]
        self.__server_version[3] = server_version[3]

        # NatNet Version info
        nnsvs = struct.unpack( 'BBBB', data[offset:offset+4] )
        offset += 4
        self.__nat_net_stream_version_server[0]=nnsvs[0]
        self.__nat_net_stream_version_server[1]=nnsvs[1]
        self.__nat_net_stream_version_server[2]=nnsvs[2]
        self.__nat_net_stream_version_server[3]=nnsvs[3]
        if (self.__nat_net_requested_version[0] == 0) and\
           (self.__nat_net_requested_version[1] == 0):
            self.__nat_net_requested_version[0] = self.__nat_net_stream_version_server[0]
            self.__nat_net_requested_version[1] = self.__nat_net_stream_version_server[1]
            self.__nat_net_requested_version[2] = self.__nat_net_stream_version_server[2]
            self.__nat_net_requested_version[3] = self.__nat_net_stream_version_server[3]
            # Determine if the bitstream version can be changed
            if (self.__nat_net_stream_version_server[0] >= 4) and (self.use_multicast == False):
                self.__can_change_bitstream_version = True



        trace_mf("Sending Application Name: ", self.__application_name)
        trace_mf("NatNetVersion " , str(self.__nat_net_stream_version_server[0]), " "
            , str(self.__nat_net_stream_version_server[1]), " "
            , str(self.__nat_net_stream_version_server[2]), " "
                , str(self.__nat_net_stream_version_server[3]))

        trace_mf("ServerVersion " , str(self.__server_version[0]), " "
            , str(self.__server_version[1]), " "
            , str(self.__server_version[2]), " "
                , str(self.__server_version[3]) )
        return offset

    def __command_thread_function( self, in_socket, stop, gprint_level):
        message_id_dict={}
        if not self.use_multicast:
            in_socket.settimeout(2.0)
        data=bytearray(0)
        # 64k buffer size
        recv_buffer_size=64*1024
        while not stop():
            # Block for input
            try:
                data, addr = in_socket.recvfrom( recv_buffer_size )
            except socket.error as msg:
                if stop():
                    #print("ERROR: command socket access error occurred:\n  %s" %msg)
                    #return 1
                    print("shutting down")
            except  socket.herror:
                print("ERROR: command socket access herror occurred")
                return 2
            except  socket.gaierror:
                print("ERROR: command socket access gaierror occurred")
                return 3
            except  socket.timeout:
                if(self.use_multicast):
                    print("ERROR: command socket access timeout occurred. Server not responding")
                    #return 4

            if len( data ) > 0 :
                #peek ahead at message_id
                message_id = get_message_id(data)
                tmp_str="mi_%1.1d"%message_id
                if tmp_str not in message_id_dict:
                    message_id_dict[tmp_str]=0
                message_id_dict[tmp_str] += 1
                
                print_level = gprint_level()
                if message_id == self.NAT_FRAMEOFDATA:
                    if print_level > 0:
                        if (message_id_dict[tmp_str] % print_level) == 0:
                            print_level = 1
                        else:
                            print_level = 0
                message_id = self.__process_message( data , print_level)

                data=bytearray(0)

            if not self.use_multicast:
                if not stop():
                    self.send_keep_alive(in_socket, self.server_ip_address, self.command_port)
        return 0

    def __data_thread_function( self, in_socket, stop, gprint_level):
        message_id_dict={}
        data=bytearray(0)
        # 64k buffer size
        recv_buffer_size=64*1024

        while not stop():
            # Block for input
            try:
                data, addr = in_socket.recvfrom( recv_buffer_size )
            except socket.error as msg:
                if not stop():
                    print("ERROR: data socket access error occurred:\n  %s" %msg)
                    return 1
            except  socket.herror:
                print("ERROR: data socket access herror occurred")
                #return 2
            except  socket.gaierror:
                print("ERROR: data socket access gaierror occurred")
                #return 3
            except  socket.timeout:
                #if self.use_multicast:
                print("ERROR: data socket access timeout occurred. Server not responding")
                #return 4
            if len( data ) > 0 :
                #peek ahead at message_id
                message_id = get_message_id(data)
                tmp_str="mi_%1.1d"%message_id
                if tmp_str not in message_id_dict:
                    message_id_dict[tmp_str]=0
                message_id_dict[tmp_str] += 1
                
                print_level = gprint_level()
                if message_id == self.NAT_FRAMEOFDATA:
                    if print_level > 0:
                        if (message_id_dict[tmp_str] % print_level) == 0:
                            print_level = 1
                        else:
                            print_level = 0
                message_id = self.__process_message( data , print_level)

                data=bytearray(0)
        return 0

    def __process_message( self, data : bytes, print_level=0):
        #return message ID
        major = self.get_major()
        minor = self.get_minor()

        trace( "Begin Packet\n-----------------" )
        show_nat_net_version = False
        if show_nat_net_version:
            trace("NatNetVersion " , str(self.__nat_net_requested_version[0]), " "\
                , str(self.__nat_net_requested_version[1]), " "\
                , str(self.__nat_net_requested_version[2]), " "\
                , str(self.__nat_net_requested_version[3]))

        message_id = get_message_id(data)

        packet_size = int.from_bytes( data[2:4], byteorder='little' )

        #skip the 4 bytes for message ID and packet_size
        offset = 4
        if message_id == self.NAT_FRAMEOFDATA :
            trace( "Message ID  : %3.1d NAT_FRAMEOFDATA"% message_id )
            trace( "Packet Size : ", packet_size )

            offset_tmp, mocap_data = self.__unpack_mocap_data( data[offset:], packet_size, major, minor )
            offset += offset_tmp
            trace("MoCap Frame: %d\n"%(mocap_data.prefix_data.frame_number))
            if print_level >= 1:
                # get a string version of the data for output
                mocap_data_str=mocap_data.get_as_string()
                print("%s\n"%mocap_data_str)

        elif message_id == self.NAT_MODELDEF :
            trace( "Message ID  : %3.1d NAT_MODELDEF"% message_id )
            trace( "Packet Size : %d"% packet_size )
            offset_tmp, data_descs = self.__unpack_data_descriptions( data[offset:], packet_size, major, minor)
            offset += offset_tmp
            trace("Data Descriptions:\n")
            if print_level>0:
                # get a string version of the data for output
                data_descs_str=data_descs.get_as_string()
                print("%s\n"%(data_descs_str))

        elif message_id == self.NAT_SERVERINFO :
            trace( "Message ID  : %3.1d NAT_SERVERINFO"% message_id )
            trace( "Packet Size : ", packet_size )
            offset += self.__unpack_server_info( data[offset:], packet_size, major, minor)

        elif message_id == self.NAT_RESPONSE :
            trace( "Message ID  : %3.1d NAT_RESPONSE"% message_id )
            trace( "Packet Size : ", packet_size )
            if packet_size == 4 :
                command_response = int.from_bytes( data[offset:offset+4], byteorder='little' )
                offset += 4
                trace( "Command response: %d"% command_response )
            else:
                show_remainder = False
                message, separator, remainder = bytes(data[offset:]).partition( b'\0' )
                offset += len( message ) + 1
                if(show_remainder):
                    trace( "Command response:", message.decode( 'utf-8' ),\
                        " separator:", separator, " remainder:",remainder )
                else:
                    trace( "Command response:", message.decode( 'utf-8' ))
        elif message_id == self.NAT_UNRECOGNIZED_REQUEST :
            trace( "Message ID  : %3.1d NAT_UNRECOGNIZED_REQUEST: "% message_id )
            trace( "Packet Size : ", packet_size )
            trace( "Received 'Unrecognized request' from server" )
        elif message_id == self.NAT_MESSAGESTRING :
            trace( "Message ID  : %3.1d NAT_MESSAGESTRING"% message_id)
            trace( "Packet Size : ", packet_size )
            message, separator, remainder = bytes(data[offset:]).partition( b'\0' )
            offset += len( message ) + 1
            trace( "Received message from server:", message.decode( 'utf-8' ) )
        else:
            trace( "Message ID  : %3.1d UNKNOWN"% message_id )
            trace( "Packet Size : ", packet_size )
            trace( "ERROR: Unrecognized packet type" )

        trace( "End Packet\n-----------------" )
        return message_id

    def send_request( self, in_socket, command, command_str, address ):
        # Compose the message in our known message format
        packet_size = 0
        if command == self.NAT_REQUEST_MODELDEF or command == self.NAT_REQUEST_FRAMEOFDATA :
            packet_size = 0
            command_str = ""
        elif command == self.NAT_REQUEST :
            packet_size = len( command_str ) + 1
        elif command == self.NAT_CONNECT :
            command_str = "Ping"
            packet_size = len( command_str ) + 1
        elif command == self.NAT_KEEPALIVE:
            packet_size = 0
            command_str = ""

        data = command.to_bytes( 2, byteorder='little' )
        data += packet_size.to_bytes( 2, byteorder='little' )

        data += command_str.encode( 'utf-8' )
        data += b'\0'

        return in_socket.sendto( data, address )

    def send_command( self, command_str):
        nTries = 3
        ret_val = -1
        while nTries:
            nTries -= 1
            ret_val = self.send_request( self.command_socket, self.NAT_REQUEST, command_str,  (self.server_ip_address, self.command_port) )
            if (ret_val != -1):
                break;
        return ret_val

        #return self.send_request(self.data_socket,    self.NAT_REQUEST, command_str,  (self.server_ip_address, self.command_port) )

    def send_commands(self,tmpCommands, print_results: bool =True):
        for sz_command in tmpCommands:
            return_code = self.send_command(sz_command)
            if(print_results):
                print("Command: %s - return_code: %d"% (sz_command, return_code) )

    def send_keep_alive(self,in_socket, server_ip_address, server_port):
        return self.send_request(in_socket, self.NAT_KEEPALIVE, "", (server_ip_address, server_port))

    def get_command_port(self):
        return self.command_port

    def get_application_name(self):
        return self.__application_name

    def get_nat_net_requested_version(self):
        return self.__nat_net_requested_version

    def get_nat_net_version_server(self):
        return self.__nat_net_stream_version_server

    def get_server_version(self):
        return self.__server_version

    def process_data_and_cmd(self):
        while True: # process as many data as possible
            try:
                data = self.data_socket.recv(64*1024)
            except socket.error as err:
                if err.errno != 10035: # WSAEWOULDBLOCK 
                    print('optitrack', err)
                break;
            else:
                if len(data) > 0:
                    self.__process_message(data)
                else:
                    break
        while True:
            try:
                data = self.command_socket.recv(64*1024)
            except socket.error as err:
                if err.errno != 10035: # WSAEWOULDBLOCK 
                    print('optitrack', err)
                break;
            else:
                if len(data) > 0:
                    self.__process_message(data)
                else:
                    break
        
    def setup_sdk(self):
        # Create the data socket
        self.data_socket = self.__create_data_socket(self.data_port)
        self.data_socket.setblocking(False)
        if self.data_socket is None :
            print( "Could not open data channel" )
            return False

        # Create the command socket
        self.command_socket = self.__create_command_socket()
        self.command_socket.setblocking(False)
        if self.command_socket is None :
            print( "Could not open command channel" )
            return False

        # Required for setup
        # Get NatNet and server versions
        self.send_request(self.command_socket, self.NAT_CONNECT, "",  (self.server_ip_address, self.command_port) )



#!/usr/bin/env python

"""
  MAVProxy default console
"""
import sys

class SimpleConsole():
    '''
    a message console for MAVProxy
    '''
    def __init__(self):
        pass

    def write(self, text, fg='black', bg='white'):
        '''write to the console'''
        if isinstance(text, str):
            sys.stdout.write(text)
        else:
            sys.stdout.write(str(text))
        sys.stdout.flush()

    def writeln(self, text, fg='black', bg='white'):
        '''write to the console with linefeed'''
        if not isinstance(text, str):
            text = str(text)
        self.write(text + '\n', fg=fg, bg=bg)

    def set_status(self, name, text='', row=0, fg='black', bg='white'):
        '''set a status value'''
        pass

    def error(self, text, fg='red', bg='white'):
        self.writeln(text, fg=fg, bg=bg)

    def close(self):
        pass

    def is_alive(self):
        '''check if we are alive'''
        return True

    def set_menu(self, menu, callback):
        pass

if __name__ == "__main__":
    # test the console
    import time
    console = SimpleConsole()
    while console.is_alive():
        console.write('Tick', fg='red')
        console.write(" %s " % time.asctime())
        console.writeln('tock', bg='yellow')
        time.sleep(0.5)



# Copyright (C) 2016  Intel Corporation. All rights reserved.
#
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Module for parsing Wavefront objects. Currently, only a small subset of the
directives are supported.

The directives supported by the object parser are:
    - v
    - vn
    - f
    - mtlib
    - usemtl

The directives supported by the material library parser are:
    - newmtl
    - Ka r g b
    - Kd r g b
    - Ks r g b
    - Ns

Unsupported directives found while parsing are stored in the parser's
ignored_directives attribute.
'''
from __future__ import print_function

import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
import threading
import sys

class Parser(object):
    def __init__(self, filename=None, string='', enable_cache=False):
        self.filename = filename
        self.string = string
        self.enable_cache = enable_cache
        self.deps = []

    def parse(self, progress_callback=None):
        if self.filename:
            return self.parse_file(progress_callback)
        return self.parse_str(progress_callback)

    def from_cache(self, progress_callback=None):
        deps_filename = self.filename + '.deps'
        if not os.path.exists(deps_filename):
            return None

        with open(deps_filename, 'r') as f:
            for line in f:
                filename = line.strip()
                if not os.path.exists(filename):
                    return None

                cache_filename = filename + '.cache'
                if not os.path.exists(cache_filename):
                    return None

                s = os.stat(filename)
                sc = os.stat(cache_filename)
                if sc.st_mtime < s.st_mtime:
                    return None

        if progress_callback:
            progress_callback(-1, -1)

        with open(self.filename + '.cache', 'r') as f:
            try:
                obj = pickle.load(f)
            except:
                print("wavefront parser: error on loading cache, falling back to parsing", file=sys.stderr)
                return None
            else:
                return obj

    def parse_file(self, progress_callback=None):
        if self.enable_cache:
            obj = self.from_cache(progress_callback=progress_callback)
            if obj:
                return obj

        s = os.stat(self.filename)
        try:
            blksize = s.st_blksize
        except AttributeError:
            blksize = 4096

        num_lines = None
        if progress_callback:
            num_lines = 0
            with open(self.filename, 'r') as f:
                try:
                    read = f.raw.read
                except AttributeError:
                    read = f.read
                blk = read(blksize)
                while blk:
                    num_lines += blk.count("\n")
                    blk = read(blksize)

        self.deps = [self.filename]

        with open(self.filename, 'r') as f:
            obj = self.parse_lines(
                f,
                num_lines=num_lines,
                progress_callback=progress_callback,
            )

            if self.enable_cache:
                # FIXME: this will make the cache files be located at the same
                # location as the source file, which might be somewhere the
                # user doesn't have write access.
                with open(self.filename + '.deps', 'w') as deps_file:
                    for filename in self.deps:
                        print(filename, file=deps_file)
                with open(self.filename + '.cache', 'w') as cache_file:
                    pickle.dump(obj, cache_file)

            return obj

    def parse_str(self, progress_callback=None):
        self.deps = []
        lines = self.string.splitlines()
        return self.parse_lines(
            lines,
            num_lines=len(lines) if progress_callback else None,
            progress_callback=progress_callback,
        )

    def parse_lines(self, lines, num_lines=None, progress_callback=None):
        obj = self.new_target()
        self.reset()

        i = 0
        for line in lines:
            line = self.filter_line(line)
            i += 1
            if not line:
                continue
            self.parse_line(line, obj)
            if progress_callback and num_lines:
                progress_callback(i, num_lines)
        return obj

    def filter_line(self, line):
        i = line.find('#')
        if i != -1:
            line = line[:i]
        return line.strip()

class ParserWorker(threading.Thread):
    def __init__(self, parser, progress_callback=None, complete_callback=None):
        super(ParserWorker, self).__init__()
        self.lock = threading.Lock()
        self.parser = parser
        self.complete_callback = complete_callback
        self.progress = 0, 0

    def run(self):
        self.obj = self.parser.parse(progress_callback=self.progress_callback)
        if self.complete_callback:
            self.complete_callback()

    def progress_callback(self, i, num_lines):
        self.lock.acquire()
        self.progress = i, num_lines
        self.lock.release()

    def get_progress(self):
        self.lock.acquire()
        p = self.progress
        self.lock.release()
        return p

class Mtl:
    def __init__(self):
        self.name = None
        self.Ka = 1.0, 1.0, 1.0
        self.Kd = 1.0, 1.0, 1.0
        self.Ks = 1.0, 1.0, 1.0
        self.Ns = 1.0

class Obj:
    def __init__(self):
        self.vertices = []
        self.normals = []
        self.faces = []
        self.materials = {}

class ObjParser(Parser):
    def reset(self):
        self.ignored_directives = set()
        self.current_mtl = Mtl()
        self.mtl_map = {}

    def new_target(self):
        return Obj()

    def parse_line(self, line, obj):
        def parse_vertex_data_ref(v):
            if not v:
                return 0
            try:
                v = int(v)
            except ValueError:
                raise Exception("vertex data references must be integers")
            if not v:
                raise Exception("vertex data references can not be zero")
            return v

        split = line.split()
        directive = split[0]
        args = split[1:]
        n = len(args)
        if directive == 'v':
            if n == 3:
                x, y, z = args
                w = 1.0
            elif n == 4:
                x, y, z, w = args
            else:
                raise Exception("wrong number of arguments for directive v")
            try:
                obj.vertices.append((float(x), float(y), float(z), float(w)))
            except ValueError:
                raise Exception("arguments for directive v must be floating point numbers")
        elif directive == 'vn':
            if n == 3:
                x, y, z = args
            else:
                raise Exception("wrong number of arguments for directive vn")
            try:
                obj.normals.append((float(x), float(y), float(z)))
            except ValueError:
                raise Exception("arguments for directive vn must be floating point numbers")
        elif directive == 'f':
            if n < 3:
                raise Exception("directive f requires at least 3 vertices")
            vertex_data = []
            for s in args:
                s = s.split('/')
                if len(s) != 3:
                    raise Exception("invalid number of references")
                v, t, n = s
                v = parse_vertex_data_ref(v)
                t = parse_vertex_data_ref(t)
                n = parse_vertex_data_ref(n)
                # TODO: address illegal statements like "f 1//1 2/2/2"
                vertex_data.append((v, t, n))
            obj.faces.append((vertex_data, self.current_mtl))
        elif directive == 'mtllib':
            if n < 1:
                raise Exception("wrong number of arguments for directive mtllib")
            d = os.path.dirname(self.filename)
            for filename in args:
                parser = MtlParser(filename=os.path.join(d, filename))
                l = parser.parse()
                self.deps += parser.deps
                for mtl in l:
                    if mtl.name in self.mtl_map:
                        continue
                    self.mtl_map[mtl.name] = mtl
        elif directive == 'usemtl':
            if n != 1:
                raise Exception("wrong number of arguments for directive usemtl")
            name = args[0]
            if name not in self.mtl_map:
                raise Exception("material %s not found" % name)
            if name not in obj.materials:
                obj.materials[name] = self.mtl_map[name]
            self.current_mtl = obj.materials[name]
        else:
            self.ignored_directives.add(directive)

class MtlParser(Parser):
    def reset(self):
        self.current_mtl = None
        self.ignored_directives = set()

    def new_target(self):
        return []

    def parse_line(self, line, mtl_list):
        def rgb():
            try:
                if n == 1:
                    r = float(args[0])
                    g = r
                    b = r
                elif n == 3:
                    r, g, b = float(args[0]), float(args[1]), float(args[2])
                else:
                    raise Exception("wrong number of arguments for directive %s" % directive)
            except ValueError:
                raise Exception("arguments for directive %s must be floating point numbers" % directive)
            else:
                return r, g, b

        def ignore_unsupported_color_statement():
           if n > 0 and args[0] in ('spectral', 'xyz'):
               self.ignored_directives.add('%s %s' % (directive, arg[0]))
               return True
           return False

        split = line.split()
        directive = split[0]
        args = split[1:]
        n = len(args)
        if directive == 'newmtl':
            if n != 1:
                raise Exception("wrong number of arguments to directive newmtl")
            self.current_mtl = Mtl()
            self.current_mtl.name = args[0]
            mtl_list.append(self.current_mtl)
        else:
            if not self.current_mtl:
                raise Exception("create a new material with newmtl before any other statement")

        if directive == 'Ka':
            if ignore_unsupported_color_statement():
                return
            self.current_mtl.Ka = rgb()
        elif directive == 'Kd':
            if ignore_unsupported_color_statement():
                return
            self.current_mtl.Kd = rgb()
        elif directive == 'Ks':
            if ignore_unsupported_color_statement():
                return
            self.current_mtl.Ks = rgb()
        elif directive == 'Ns':
            if n != 1:
                raise Exception("wrong number of arguments for directive Ns")
            try:
                self.current_mtl.Ns = float(args[0])
            except ValueError:
                raise Exception("argument for directive Ns must be a floating point number")
        else:
            self.ignored_directives.add(directive)



'''
Swarming support
Stephen Dade, July 2021

Helper GUI for managing leader-follower swarms

'''

import sys
import time
import wx.lib.scrolledpanel as scrolled
from pymavlink import mavutil

from MAVProxy.modules.lib import (icon, mp_module, mp_settings, mp_util,
                                  multiproc, win_layout)
from MAVProxy.modules.lib.wx_loader import wx



def get_vehicle_name(vehtype):
    '''return vehicle type string from a heartbeat'''
    if vehtype == mavutil.mavlink.MAV_TYPE_FIXED_WING:
        return 'Plane'
    if vehtype == mavutil.mavlink.MAV_TYPE_GROUND_ROVER:
        return 'Rover'
    if vehtype == mavutil.mavlink.MAV_TYPE_SURFACE_BOAT:
        return 'Boat'
    if vehtype == mavutil.mavlink.MAV_TYPE_SUBMARINE:
        return 'Sub'
    if vehtype in [mavutil.mavlink.MAV_TYPE_QUADROTOR,
                   mavutil.mavlink.MAV_TYPE_COAXIAL,
                   mavutil.mavlink.MAV_TYPE_HEXAROTOR,
                   mavutil.mavlink.MAV_TYPE_OCTOROTOR,
                   mavutil.mavlink.MAV_TYPE_TRICOPTER,
                   mavutil.mavlink.MAV_TYPE_DODECAROTOR]:
        return "Copter"
    if vehtype == mavutil.mavlink.MAV_TYPE_HELICOPTER:
        return "Heli"
    if vehtype == mavutil.mavlink.MAV_TYPE_ANTENNA_TRACKER:
        return "Tracker"
    if vehtype == mavutil.mavlink.MAV_TYPE_AIRSHIP:
        return "Blimp"
    return "UNKNOWN({0})".format(vehtype)

def parmString(str):
    '''convert string to bytes in python3'''
    if sys.version_info.major < 3:
        return str   
    return bytes(str, 'ascii')

class SwarmUI():
    '''
    Swarm User Interface
    '''

    def __init__(self, parmsToShow, title='MAVProxy: Swarm Control', takeoffalt=10):
        self.title = title
        self.parmsToShow = parmsToShow
        self.takeoffalt = takeoffalt
        self.parent_pipe, self.child_pipe = multiproc.Pipe()
        self.close_event = multiproc.Event()
        self.close_event.clear()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        mp_util.child_close_fds()

        app = wx.App(False)
        app.frame = SwarmFrame(
            self, title=self.title, params=self.parmsToShow, takeoffalt=self.takeoffalt)
        app.frame.Show()
        app.MainLoop()

    def changesetting(self, setting, val):
        '''Change a setting'''
        if self.child.is_alive():
            self.parent_pipe.send(('changesetting', (setting, val)))

    def close(self):
        '''close the UI'''
        self.close_event.set()
        if self.is_alive():
            self.child.join(2)

    def is_alive(self):
        '''check if child is still going'''
        return self.child.is_alive()

    def updateLayout(self, newLayout):
        '''set a status value'''
        if self.child.is_alive():
            self.parent_pipe.send(('updatelayout', newLayout))

    def updateHB(self, vehHB):
        '''update HB status of vehicles'''
        if self.child.is_alive():
            self.parent_pipe.send(('updateHB', vehHB))

    def onmavlinkpacket(self, pkt):
        '''set a status value'''
        if self.child.is_alive():
            self.parent_pipe.send(('onmavlinkpacket', pkt))


class UnassignedPanel(wx.Panel):
    '''
    A wx.Panel to hold a single unassigned vehicle
    '''
    def __init__(self, state, parent):
        wx.Panel.__init__(self, parent)  # , style=wx.BORDER_SIMPLE)
        self.state = state

        self.sizer = wx.BoxSizer(wx.VERTICAL)

        self.titleText = wx.StaticText(
            self, label="Unassigned Vehicles -->\n\nTools:")
        self.sizer.Add(self.titleText, flag=wx.EXPAND |
                       wx.LEFT | wx.RIGHT, border=5)

        self.getParamsButton = wx.Button(
            self, label="Get offsets", size=wx.Size(100, 50))
        self.Bind(wx.EVT_BUTTON, self.getParams, self.getParamsButton)
        self.resetButton = wx.Button(
            self, label="Reset layout", size=wx.Size(100, 50))
        self.Bind(wx.EVT_BUTTON, self.reset, self.resetButton)

        self.sizer.Add(self.getParamsButton)
        self.sizer.Add(self.resetButton)

        # Do the sizer layout
        self.SetSizer(self.sizer)

    def getParams(self, event):
        '''get offset params again'''
        self.state.child_pipe.send(("getparams", None, None))

    def reset(self, event):
        '''arm/disarm the vehicle'''
        self.state.child_pipe.send(("resetLayout", None, None))


class VehiclePanel(wx.Panel):
    '''
    A wx.panel to hold a single vehicle (leader or follower)
    '''

    def __init__(self, state, parent, sysid, compid, vehtype, isLeader, listFollowers, takeoffalt):
        wx.Panel.__init__(self, parent)
        self.state = state
        self.sysid = sysid
        self.compid = compid
        self.listFollowers = listFollowers
        self.isLeader = isLeader
        self.vehtype = vehtype

        self.takeoffalt = takeoffalt

        self.inLinkLoss = False

        # XYZ offsets. Filled are we get params
        self.offsetValues = [None, None, None]

        self.sizer = wx.BoxSizer(wx.VERTICAL)

        # Status boxes
        if self.isLeader:
            self.title = wx.StaticText(self, label="Leader {0}:{1} {2}".format(
                sysid, compid, get_vehicle_name(vehtype)))
            self.offsets = wx.StaticText(self, label="Offset: N/A")
        else:
            self.title = wx.StaticText(self, label="Veh {0}:{1} {2}".format(
                sysid, compid, get_vehicle_name(vehtype)))
            self.offsets = wx.StaticText(self, label="Offset: xxxxxx")

        self.armmode = wx.StaticText(self, label="armed/mode N/A    ")
        self.thrAlt = wx.StaticText(
            self, label="Alt: {0}m    Thr: {1}%".format(0, 0))
        self.altRel = wx.StaticText(self, label="Rel Alt: {0}m".format(0))
        self.battery = wx.StaticText(self, label="Battery: {0}V".format(0))
        self.status = wx.StaticText(self, label="Status: N/A")
        self.prearm = wx.StaticText(self, label="Prearm: N/A")
        self.statusText = wx.TextCtrl(
            self, style=wx.TE_READONLY | wx.TE_MULTILINE, size=wx.Size(140, 100))

        # Command buttons
        self.doArm = wx.Button(self, label="XXX", size=wx.Size(100, 50))
        self.Bind(wx.EVT_BUTTON, self.arm, self.doArm)
        if self.isLeader:
            self.armSizer = wx.BoxSizer(wx.HORIZONTAL)
            self.doArmAll = wx.Button(self, label="ALL", size=wx.Size(70, 50))
            self.Bind(wx.EVT_BUTTON, self.armAll, self.doArmAll)

        self.doGuided = wx.Button(
            self, label="Mode GUIDED", size=wx.Size(100, 50))
        self.Bind(wx.EVT_BUTTON, self.guided, self.doGuided)
        if self.isLeader:
            self.guidedSizer = wx.BoxSizer(wx.HORIZONTAL)
            self.doGuidedAll = wx.Button(
                self, label="ALL", size=wx.Size(70, 50))
            self.Bind(wx.EVT_BUTTON, self.guidedAll, self.doGuidedAll)

        if vehtype != mavutil.mavlink.MAV_TYPE_GROUND_ROVER:
            self.doGuidedTakeoff = wx.Button(
                self, label="GUIDED T/O {0}m".format(self.takeoffalt), size=wx.Size(130, 50))
            self.Bind(wx.EVT_BUTTON, self.guidedTakeoff, self.doGuidedTakeoff)
            if self.isLeader:
                self.takeoffSizer = wx.BoxSizer(wx.HORIZONTAL)
                self.doGuidedTakeoffAll = wx.Button(
                    self, label="ALL".format(self.takeoffalt), size=wx.Size(70, 50))
                self.Bind(wx.EVT_BUTTON, self.guidedTakeoffAll,
                          self.doGuidedTakeoffAll)

        self.doRTL = wx.Button(self, label="Mode RTL", size=wx.Size(100, 50))
        self.Bind(wx.EVT_BUTTON, self.rtl, self.doRTL)
        if self.isLeader:
            self.rtlSizer = wx.BoxSizer(wx.HORIZONTAL)
            self.doRTLAll = wx.Button(self, label="ALL", size=wx.Size(70, 50))
            self.Bind(wx.EVT_BUTTON, self.rtlAll, self.doRTLAll)

        self.doKill = wx.Button(self, label="KILL", size=wx.Size(100, 50))
        self.killTimer = None
        self.Bind(wx.EVT_BUTTON, self.kill, self.doKill)
        if self.isLeader:
            self.killSizer = wx.BoxSizer(wx.HORIZONTAL)
            self.doKillAll = wx.Button(self, label="ALL", size=wx.Size(70, 50))
            self.killAllTimer = None
            self.Bind(wx.EVT_BUTTON, self.killall, self.doKillAll)

        if self.isLeader:
            self.doFollowAll = wx.Button(
                self, label="All Follow Leader", size=wx.Size(130, 50))
            self.Bind(wx.EVT_BUTTON, self.followAll, self.doFollowAll)
            self.doAUTO = wx.Button(
                self, label="Mode AUTO", size=wx.Size(100, 50))
            self.Bind(wx.EVT_BUTTON, self.auto, self.doAUTO)
        else:
            self.doFollow = wx.Button(
                self, label="Mode Follow", size=wx.Size(100, 50))
            self.Bind(wx.EVT_BUTTON, self.follow, self.doFollow)

        # Do the sizer layout
        self.doSizer()

        # get offset params. Needs to be after GUI elements are created
        time.sleep(0.05)
        self.state.child_pipe.send(("getoffsets", self.sysid, self.compid))

    def doSizer(self):
        '''Sort out all the sizers and layout'''
        self.sizer.Add(self.title)
        self.sizer.Add(self.armmode)
        self.sizer.Add(self.thrAlt)
        self.sizer.Add(self.altRel)
        self.sizer.Add(self.battery)
        self.sizer.Add(self.status)
        self.sizer.Add(self.prearm)
        self.sizer.Add(self.offsets)
        self.sizer.Add(self.statusText)

        if self.isLeader:
            self.armSizer.Add(self.doArm)
            self.armSizer.Add(self.doArmAll)
            self.sizer.Add(self.armSizer)
        else:
            self.sizer.Add(self.doArm)

        if self.vehtype != mavutil.mavlink.MAV_TYPE_GROUND_ROVER:
            if self.isLeader:
                self.takeoffSizer.Add(self.doGuidedTakeoff)
                self.takeoffSizer.Add(self.doGuidedTakeoffAll)
                self.sizer.Add(self.takeoffSizer)
            else:
                self.sizer.Add(self.doGuidedTakeoff)

        if self.isLeader:
            self.guidedSizer.Add(self.doGuided)
            self.guidedSizer.Add(self.doGuidedAll)
            self.sizer.Add(self.guidedSizer)
        else:
            self.sizer.Add(self.doGuided)

        if self.isLeader:
            self.rtlSizer.Add(self.doRTL)
            self.rtlSizer.Add(self.doRTLAll)
            self.sizer.Add(self.rtlSizer)
        else:
            self.sizer.Add(self.doRTL)

        if self.isLeader:
            self.killSizer.Add(self.doKill)
            self.killSizer.Add(self.doKillAll)
            self.sizer.Add(self.killSizer)
        else:
            self.sizer.Add(self.doKill)

        if self.isLeader:
            self.sizer.Add(self.doFollowAll)
            self.sizer.Add(self.doAUTO)
        else:
            self.sizer.Add(self.doFollow)

        self.SetSizer(self.sizer)

    def updatetakeoffalt(self, alt):
        '''update takeoff altitude'''
        self.takeoffalt = alt
        if self.vehtype != mavutil.mavlink.MAV_TYPE_GROUND_ROVER:
            self.doGuidedTakeoff.SetLabel(
                "GUIDED T/O {0}m".format(self.takeoffalt))
            if self.isLeader:
                self.doGuidedTakeoffAll.SetLabel("ALL".format(self.takeoffalt))

    def updateData(self, armed, mode):
        '''update the arming/mode status. Only change if required'''
        
        if armed and (self.doArm.GetLabel() in ["ARM", "XXX"] or str(mode) not in self.armmode.GetLabel()):
            self.armmode.SetForegroundColour((0, 200, 0))
            self.armmode.SetLabel("Armed" + "/" + str(mode))
            self.doArm.SetLabel("DISARM")
        elif not armed and (self.doArm.GetLabel() in ["DISARM", "XXX"] or str(mode) not in self.armmode.GetLabel()):
            self.armmode.SetForegroundColour((200, 0, 0))
            self.armmode.SetLabel("Disarmed" + "/" + str(mode))
            self.doArm.SetLabel("ARM")

    def updatethralt(self, throttle, alt):
        '''update throttle and altitude status'''
        self.thrAlt.SetLabel("Alt: {0}m    Thr: {1}%".format(int(alt), throttle))

    def updaterelalt(self, relalt):
        '''update relative altitude'''
        self.altRel.SetLabel("Rel Alt: {0}m".format(int(relalt)))

    def updateStatus(self, status):
        '''update status'''
        self.status.SetLabel("Status: {0}".format(status))
        if status in ["FAILSAFE", "Emergency", "Terminated"]:
            self.status.SetForegroundColour((200, 0, 0))
        elif status == "Flying":
            self.status.SetForegroundColour((0, 200, 0))
        else:
            self.status.SetForegroundColour((120, 120, 0))

    def updateprearm(self, preArmGood):
        '''Update pre-arm status'''
        self.prearm.SetLabel("Prearm: {0}".format(
            "OK" if preArmGood else "BAD"))
        if preArmGood:
            self.prearm.SetForegroundColour((0, 200, 0))
        else:
            self.prearm.SetForegroundColour((200, 0, 0))

    def updatevoltage(self, voltage):
        '''update battery voltage'''
        self.battery.SetLabel("Battery: {0:.1f}V".format(voltage))

    def arm(self, event):
        '''arm/disarm the vehicle'''
        self.state.child_pipe.send(
            ("arm" if self.doArm.GetLabel() == "ARM" else "disarm", self.sysid, self.compid))

    def armAll(self, event):
        '''Arm leader and all followers'''
        self.state.child_pipe.send(("arm", self.sysid, self.compid))
        if len(self.listFollowers) > 0:
            for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                self.state.child_pipe.send(("arm", sysidFollow, compidFollow))

    def guided(self, event):
        '''switch to mode guided'''
        self.state.child_pipe.send(("GUIDED", self.sysid, self.compid))

    def guidedAll(self, event):
        '''mode guided for leader and all followers'''
        self.state.child_pipe.send(("GUIDED", self.sysid, self.compid))
        if len(self.listFollowers) > 0:
            for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                self.state.child_pipe.send(
                    ("GUIDED", sysidFollow, compidFollow))

    def addstatustext(self, text):
        '''have recieved statustext, update relevant control'''
        self.statusText.AppendText('\n' + text)

    def guidedTakeoff(self, event):
        '''switch to guided mode, wait 0.1 sec then nav_takeoff to TAKEOFFALT'''
        self.state.child_pipe.send(("GUIDED", self.sysid, self.compid))
        time.sleep(0.1)
        self.state.child_pipe.send(("takeoff", self.sysid, self.compid))

    def guidedTakeoffAll(self, event):
        '''switch to guided mode, wait 0.1 sec then nav_takeoff to TAKEOFFALT for all vehicles'''
        self.state.child_pipe.send(("GUIDED", self.sysid, self.compid))
        time.sleep(0.1)
        self.state.child_pipe.send(("takeoff", self.sysid, self.compid))

        if len(self.listFollowers) > 0:
            for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                self.state.child_pipe.send(
                    ("GUIDED", sysidFollow, compidFollow))
                time.sleep(0.1)
                self.state.child_pipe.send(
                    ("takeoff", sysidFollow, compidFollow))

    def rtl(self, event):
        '''switch to mode RTL'''
        self.state.child_pipe.send(("RTL", self.sysid, self.compid))

    def rtlAll(self, event):
        '''RTL leader and all followers'''
        self.state.child_pipe.send(("RTL", self.sysid, self.compid))
        if len(self.listFollowers) > 0:
            for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                self.state.child_pipe.send(("RTL", sysidFollow, compidFollow))

    def killTimeout(self, event):
        '''Kill action timeout'''
        self.doKill.SetLabel("KILL")
        self.killTimer = None

    def killAllTimeout(self, event):
        '''Kill all action timeout'''
        self.doKillAll.SetLabel("ALL")
        self.killAllTimer = None

    def kill(self, event):
        '''disarm force (kill) the drone'''
        if self.doKill.GetLabel() == "KILL":
            self.doKill.SetLabel("KILL Sure?")
            # start a 1 sec timer
            self.killTimer = wx.Timer(self)
            self.Bind(wx.EVT_TIMER, self.killTimeout, self.killTimer)
            self.killTimer.Start(1000)
        else:
            self.state.child_pipe.send(("kill", self.sysid, self.compid))

    def killall(self, event):
        '''disarm force (kill) the whole swarm'''
        if self.doKillAll.GetLabel() == "ALL":
            self.doKillAll.SetLabel("ALL Sure?")
            # start a 1 sec timer
            self.killAllTimer = wx.Timer(self)
            self.Bind(wx.EVT_TIMER, self.killAllTimeout, self.killAllTimer)
            self.killAllTimer.Start(1000)
        else:
            self.state.child_pipe.send(("kill", self.sysid, self.compid))
            if len(self.listFollowers) > 0:
                for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                    self.state.child_pipe.send(
                        ("kill", sysidFollow, compidFollow))

    def follow(self, event):
        '''switch to mode follow'''
        self.state.child_pipe.send(("FOLLOW", self.sysid, self.compid))

    def followAll(self, event):
        '''Mode FOLLOW for all followers'''
        if len(self.listFollowers) > 0:
            for (sysidFollow, compidFollow, vehtype) in self.listFollowers:
                self.state.child_pipe.send(
                    ("FOLLOW", sysidFollow, compidFollow))

    def auto(self, event):
        '''switch to mode AUTO'''
        self.state.child_pipe.send(("AUTO", self.sysid, self.compid))

    def updateOffset(self, param, value):
        '''get offset param'''
        if self.isLeader:
            return
        if param == 'FOLL_OFS_X':
            self.offsetValues[0] = value
        elif param == 'FOLL_OFS_Y':
            self.offsetValues[1] = value
        elif param == 'FOLL_OFS_Z':
            self.offsetValues[2] = value
        else:
            return
        # refresh static text
        if self.offsetValues[0] != None and self.offsetValues[1] != None and self.offsetValues[2] != None:
            self.offsets.SetLabel("Offset: {0}m,{1}m,{2}m".format(
                self.offsetValues[0], self.offsetValues[1], self.offsetValues[2]))


class SwarmFrame(wx.Frame):
    '''
    The main frame of the UI. Runs in a subprocess, so need a pipe to communicate with
    '''

    def __init__(self, state, title, params, takeoffalt):
        self.state = state
        wx.Frame.__init__(self, None, title=title, size=(
            1400, 500), style=wx.DEFAULT_FRAME_STYLE)
        self.panel = scrolled.ScrolledPanel(self, style=wx.FULL_REPAINT_ON_RESIZE)
        self.panel.SetBackgroundColour(wx.WHITE)

        # Icon
        self.SetIcon(icon.SimpleIcon("Swarm").get_ico())

        # Params to show (array)
        self.parmsToShow = params

        self.takeoffalt = takeoffalt

        self.last_layout_send = time.time()

        # layout. Column per leader, with followers underneath
        self.sizer = wx.FlexGridSizer(1, 1, 0, 0)

        # add in the pipe from MAVProxy
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, lambda evt,
                  panel=self.panel: self.on_timer(evt), self.timer)
        self.timer.Start(100)

        #Fires on window resize
        self.Bind(wx.EVT_SIZE, self.OnSize)

        # layout
        self.panel.SetSizer(self.sizer)
        self.panel.Layout()
        self.panel.SetupScrolling()
        self.Show(True)

    def OnSize(self, e):
        #This refresh shouldn't be necessary
        self.Refresh()

        #Pass event up the chain so window still resizes
        e.Skip()

    def on_timer(self, event):
        '''receive messages from MAVProxy and pass on to GUI'''
        state = self.state

        # update saved pos/size of GUI
        now = time.time()
        if now - self.last_layout_send > 1:
            self.last_layout_send = now
            state.child_pipe.send(
                (win_layout.get_wx_window_layout(self), 0, 0))

        if state.close_event.wait(0.001):
            self.timer.Stop()
            self.Destroy()
            return
        while state.child_pipe.poll():
            obj = state.child_pipe.recv()
            if obj[0] == 'updatelayout':
                self.updateLayout(obj[1])
            elif obj[0] == 'onmavlinkpacket':
                self.onmavlinkpacket(obj[1])
            elif obj[0] == 'updateHB':
                self.updateHB(obj[1])
            elif isinstance(obj[0], win_layout.WinLayout):
                win_layout.set_wx_window_layout(self, obj[0])
            elif obj[0] == 'changesetting' and obj[1][0] == 'takeoffalt':
                self.updatetakeoffalt(int(obj[1][1]))

    def updatetakeoffalt(self, alt):
        '''Update the takeoff altitude of the vehicles'''
        children = self.sizer.GetChildren()
        # update panel
        for child in children:
            widget = child.GetWindow()
            if isinstance(widget, VehiclePanel):
                widget.updatetakeoffalt(alt)

    def updateHB(self, vehHB):
        '''Update the GUI panels if we've lost link to vehicle
        Panel goes red if more than 4 sec since last HB'''
        children = self.sizer.GetChildren()
        # update panel
        for child in children:
            widget = child.GetWindow()
            if isinstance(widget, VehiclePanel):
                if (widget.sysid, widget.compid) in vehHB.keys():
                    # put to red if more than 4 sec, else no colour. Only change if required.
                    if vehHB[(widget.sysid, widget.compid)] + 4 < time.time() and widget.GetBackgroundColour() != wx.RED:
                        widget.SetBackgroundColour(wx.RED)
                    elif vehHB[(widget.sysid, widget.compid)] + 4 > time.time() and widget.GetBackgroundColour() != wx.WHITE:
                        widget.SetBackgroundColour(wx.WHITE)

    def updateLayout(self, layout):
        '''Update (recreate) the GUI layout, based on known vehicles'''
        leaders = []  # array of (sysid, compid, followid, vehtype)

        # create list of leaders, another vehicle has it's sysid as leader
        for veh in layout:
            for vehL in layout:
                if vehL[2] == veh[0] and veh not in leaders:
                    leaders.append(veh)

        # create list of followers
        followers = {}  # dict of key=leadersysid, (sysid, compid, vehtype)
        maxfollowers = 0
        for veh in layout:
            # don't include leaders in the follower list
            if veh[2] != 0:
                if veh[2] not in followers.keys():
                    followers[veh[2]] = [(veh[0], veh[1], veh[3])]
                else:
                    followers[veh[2]].append((veh[0], veh[1], veh[3]))
                if len(followers[veh[2]]) > maxfollowers:
                    maxfollowers = len(followers[veh[2]])

        # sort followers by sysid increasing
        for fwr in followers.keys():
            followers[fwr].sort(key=lambda tup: tup[0])

        # Any unassigned vehicles: Leader with 0 followers OR leader not present
        # OR vehicles not present in either list (leftovers)
        unassignedVeh = []  # (sysid, compid, vehtype)

        # Leader with no followers
        for index, veh in enumerate(leaders):
            if veh[0] not in followers.keys():
                del leaders[index]
                unassignedVeh.append((veh[0], veh[1], veh[3]))

        # Not present in either list
        allfollowersflat = []
        for fl in followers.items():
            for vv in fl[1]:
                allfollowersflat.append(vv[0])
        for veh in layout:
            if (veh not in leaders) and veh[0] not in allfollowersflat:
                unassignedVeh.append((veh[0], veh[1], veh[3]))

        # Leader not present
        for leader in followers.copy().keys():
            if leader not in [i[0] for i in leaders]:
                for veh in followers[leader]:
                    unassignedVeh.append(veh)
                del followers[leader]  # followers.remove(leader)

        self.sizer.Clear(True)
        colsVeh = max(maxfollowers, len(unassignedVeh))
        self.sizer = wx.FlexGridSizer((len(leaders)*2) + 1, colsVeh+1+1, 5, 0)

        # start populating the grid
        # for each row:
        for leader in leaders:
            panelLeader = VehiclePanel(
                self.state, self.panel, leader[0], leader[1], leader[3], True, followers[leader[0]], self.takeoffalt)
            self.sizer.Add(panelLeader)

            # add vertical line
            line = wx.StaticLine(self.panel, style=wx.LI_VERTICAL)
            self.sizer.Add(line, proportion=0, flag=wx.EXPAND |
                           wx.LEFT | wx.RIGHT, border=5)

            for follower in range(colsVeh):
                if leader[0] in followers.keys() and len(followers[leader[0]]) > follower:
                    panelFollower = VehiclePanel(self.state, self.panel, followers[leader[0]][follower][0], followers[
                                                 leader[0]][follower][1], followers[leader[0]][follower][2], False, None, self.takeoffalt)
                else:
                    panelFollower = wx.StaticText(self.panel, label="N/A")
                self.sizer.Add(panelFollower, flag=wx.EXPAND |
                               wx.LEFT | wx.RIGHT, border=5)

            # add horizontal line, column by column
            for follower in range(colsVeh+1+1):
                line = wx.StaticLine(self.panel, style=wx.LI_HORIZONTAL)
                self.sizer.Add(line, proportion=0,
                               flag=wx.EXPAND | wx.ALL, border=0)

        # add "unassigned" row
        panelUnassigned = UnassignedPanel(self.state, self.panel)
        self.sizer.Add(panelUnassigned, flag=wx.EXPAND |
                       wx.LEFT | wx.RIGHT, border=5)
        line = wx.StaticLine(self.panel, style=wx.LI_VERTICAL)
        self.sizer.Add(line, proportion=0, flag=wx.EXPAND |
                       wx.LEFT | wx.RIGHT, border=5)

        for veh in unassignedVeh:
            panelunassigned = VehiclePanel(
                self.state, self.panel, veh[0], veh[1], veh[2], False, None, self.takeoffalt)
            self.sizer.Add(panelunassigned)

        self.panel.SetSizer(self.sizer)
        self.panel.Layout()
        self.panel.SetupScrolling()

    def onmavlinkpacket(self, msg):
        ''' on new mavlink packet, update relevant GUI elements'''
        children = self.sizer.GetChildren()
        sysid = msg.get_srcSystem()
        compid = msg.get_srcComponent()
        mtype = msg.get_type()

        for child in children:
            widget = child.GetWindow()
            # update arm and mode status on leader
            if isinstance(widget, VehiclePanel) and widget.sysid == sysid and widget.compid == compid:
                if mtype == 'HEARTBEAT':
                    mode_map = mavutil.mode_mapping_bynumber(msg.type)
                    isArmed = True if msg.base_mode & mavutil.mavlink.MAV_MODE_FLAG_SAFETY_ARMED == 128 else False
                    widget.updateData(isArmed, mode_map[msg.custom_mode])
                    sysStatus = msg.system_status
                    if sysStatus == mavutil.mavlink.MAV_STATE_BOOT:
                        widget.updateStatus("Booting")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_CALIBRATING:
                        widget.updateStatus("Calibrating")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_STANDBY:
                        widget.updateStatus("On Ground")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_ACTIVE:
                        widget.updateStatus("Flying")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_CRITICAL:
                        widget.updateStatus("FAILSAFE")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_EMERGENCY:
                        widget.updateStatus("Emergency")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_POWEROFF:
                        widget.updateStatus("Poweroff")
                    elif sysStatus == mavutil.mavlink.MAV_STATE_FLIGHT_TERMINATION:
                        widget.updateStatus("Terminated")
                elif mtype == 'VFR_HUD':
                    widget.updatethralt(msg.throttle, msg.alt)
                elif mtype == 'GLOBAL_POSITION_INT':
                    widget.updaterelalt(msg.relative_alt * 1.0e-3)
                elif mtype == "SYS_STATUS":
                    preArmGood = ((msg.onboard_control_sensors_health & mavutil.mavlink.MAV_SYS_STATUS_PREARM_CHECK)
                                  == mavutil.mavlink.MAV_SYS_STATUS_PREARM_CHECK)
                    widget.updateprearm(preArmGood)
                    widget.updatevoltage(float(msg.voltage_battery)/1000)
                elif mtype == "STATUSTEXT" and msg.severity <= mavutil.mavlink.MAV_SEVERITY_WARNING:
                    # Only pass on warning messages or more severe
                    widget.addstatustext(msg.text)
                elif mtype == 'PARAM_VALUE' and msg.param_id in self.parmsToShow:
                    widget.updateOffset(msg.param_id, int(msg.param_value))


class swarm(mp_module.MPModule):
    '''
    The MAVProxy module that manages the GUI
    '''
    def __init__(self, mpstate):
        '''Initialise module'''
        super(swarm, self).__init__(mpstate, "swarm",
                                    "swarm module", multi_vehicle=True)

        # array of tuples for (SYSID, COMPID, FOLL_SYSID, veh_type) of all detected vehicles
        self.vehicleListing = []

        self.add_command('swarm', self.cmd_swarm, "swarm control",
                         ["<status>", "set (SWARMSETTING)"])

        self.swarm_settings = mp_settings.MPSettings(
            [("takeoffalt", int, 10)])  # meters
        self.add_completion_function('(SWARMSETTING)',
                                     self.swarm_settings.completion)

        # Which params to show on the GUI per vehicle
        self.parmsToShow = ["FOLL_OFS_X", "FOLL_OFS_Y", "FOLL_OFS_Z"]

        # time of last HB for each vehicle. Key is tuple of sysid,compid. Value is time of last HB
        self.vehicleLastHB = {}
        self.needHBupdate_timer = mavutil.periodic_event(1)

        # Periodic event to update the GUI
        self.needGUIupdate = False
        self.needGUIupdate_timer = mavutil.periodic_event(1)

        # Periodic event to send param (offset) requests (5 Hz)
        self.requestParams_timer = mavutil.periodic_event(5)

        # Periodic event re-get params (0.1 Hz)
        self.RerequestParams_timer = mavutil.periodic_event(0.1)

        # List of any vehicles we still need to get params for
        self.vehParamsToGet = []

        # All vehicle positions. Dict. Key is sysid, value is tuple of (lat,lon,alt)
        self.allVehPos = {}

        self.validVehicles = frozenset([mavutil.mavlink.MAV_TYPE_FIXED_WING,
                              mavutil.mavlink.MAV_TYPE_GROUND_ROVER,
                              mavutil.mavlink.MAV_TYPE_SURFACE_BOAT,
                              mavutil.mavlink.MAV_TYPE_SUBMARINE,
                              mavutil.mavlink.MAV_TYPE_QUADROTOR,
                              mavutil.mavlink.MAV_TYPE_COAXIAL,
                              mavutil.mavlink.MAV_TYPE_HEXAROTOR,
                              mavutil.mavlink.MAV_TYPE_OCTOROTOR,
                              mavutil.mavlink.MAV_TYPE_TRICOPTER,
                              mavutil.mavlink.MAV_TYPE_HELICOPTER,
                              mavutil.mavlink.MAV_TYPE_DODECAROTOR,
                              mavutil.mavlink.MAV_TYPE_AIRSHIP])

        # The GUI
        self.gui = SwarmUI(
            self.parmsToShow, takeoffalt=self.swarm_settings.get('takeoffalt'))

    def cmd_swarm(self, args):
        '''swarm command parser'''
        usage = "usage: swarm <set>"
        if len(args) == 0:
            print(usage)
        elif args[0] == "set":
            self.swarm_settings.command(args[1:])
            if len(args) == 3:
                self.gui.changesetting(args[1], args[2])
        else:
            print(usage)

    def set_layout(self, layout):
        '''set window layout'''
        self.gui.parent_pipe.send([layout])

    def idle_task(self):
        '''run on idle'''
        # send updated HB stats to GUI every 1 sec
        if self.needHBupdate_timer.trigger():
            self.gui.updateHB(self.vehicleLastHB)

        # do we need to update the GUI?
        if self.needGUIupdate_timer.trigger() and self.needGUIupdate:
            self.gui.updateLayout(self.vehicleListing)
            self.needGUIupdate = False

        # do we need to get any vehicle follow sysid params?
        # only send param requests 1 per 0.1sec, to avoid link flooding
        if self.requestParams_timer.trigger() and len(self.vehParamsToGet) > 0:
            (sysid, compid) = self.vehParamsToGet.pop(0)
            self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.param_request_read_send(
                sysid, compid, parmString("FOLL_SYSID"), -1))

        if self.RerequestParams_timer.trigger():
            # If any in vehicleListing are missing their FOLL_SYSID, re-request
            for veh in self.vehicleListing:
                if veh[2] == 0:
                    self.vehParamsToGet.append(((veh[0], veh[1])))

        # execute any commands from GUI via parent_pipe
        if self.gui.parent_pipe.poll():
            (cmd, sysid, compid) = self.gui.parent_pipe.recv()
            if isinstance(cmd, win_layout.WinLayout):
                win_layout.set_layout(cmd, self.set_layout)
            if cmd == "arm":
                self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.command_long_send(
                    # self.master.mav.command_long_send(
                    sysid,  # target_system
                    compid,
                    mavutil.mavlink.MAV_CMD_COMPONENT_ARM_DISARM,  # command
                    0,  # confirmation
                    1,  # param1 (1 to indicate arm)
                    0,  # param2  (all other params meaningless)
                    0,  # param3
                    0,  # param4
                    0,  # param5
                    0,  # param6
                    0))  # param7
            elif cmd == "disarm":
                self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.command_long_send(
                    sysid,  # target_system
                    compid,
                    mavutil.mavlink.MAV_CMD_COMPONENT_ARM_DISARM,  # command
                    0,  # confirmation
                    0,  # param1 (1 to indicate arm)
                    0,  # param2  (all other params meaningless)
                    0,  # param3
                    0,  # param4
                    0,  # param5
                    0,  # param6
                    0))  # param7
            elif cmd == "takeoff":
                self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.command_long_send(
                    sysid,  # target_system
                    compid,  # target_component
                    mavutil.mavlink.MAV_CMD_NAV_TAKEOFF,  # command
                    0,  # confirmation
                    0,  # param1
                    0,  # param2
                    0,  # param3
                    0,  # param4
                    0,  # param5
                    0,  # param6
                    self.swarm_settings.takeoffalt))  # param7
            elif cmd in ["FOLLOW", "RTL", "AUTO", "GUIDED"]:
                mode_mapping = self.master.mode_mapping()
                self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.command_long_send(sysid,
                                                                                          compid,
                                                                                          mavutil.mavlink.MAV_CMD_DO_SET_MODE,
                                                                                          0,
                                                                                          mavutil.mavlink.MAV_MODE_FLAG_CUSTOM_MODE_ENABLED,
                                                                                          mode_mapping[cmd],
                                                                                          0,
                                                                                          0,
                                                                                          0,
                                                                                          0,
                                                                                          0))
            elif cmd == "kill":
                self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.command_long_send(
                    sysid,  # target_system
                    compid,
                    mavutil.mavlink.MAV_CMD_COMPONENT_ARM_DISARM,  # command
                    0,  # confirmation
                    0,  # param1 (0 to indicate disarm)
                    21196,  # param2 (indicates force disarm)
                    0,  # param3
                    0,  # param4
                    0,  # param5
                    0,  # param6
                    0))  # param7
            elif cmd == 'getoffsets':
                # time.sleep(0.1)
                for parm in self.parmsToShow:
                    self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.param_request_read_send(
                        sysid, compid, parmString(parm), -1))
            elif cmd == 'resetLayout':
                self.vehicleListing = []
                self.vehicleLastHB = {}
                self.needGUIupdate = True
                # self.gui.updateLayout(self.vehicleListing)
            elif cmd == 'getparams':
                for (sysid, compid, foll_sysid, veh_type) in self.vehicleListing:
                    # time.sleep(0.1)
                    for parm in self.parmsToShow:
                        self.mpstate.foreach_mav(sysid, compid, lambda mav: mav.param_request_read_send(
                            sysid, compid, parmString(parm), -1))

    def mavlink_packet(self, m):
        '''handle incoming mavlink packets'''
        mtype = m.get_type()
        sysid = m.get_srcSystem()
        compid = m.get_srcComponent()

        # add to GUI if vehicle not seen before
        if mtype == 'HEARTBEAT' and m.type in self.validVehicles and not (sysid in [sysidList[0] for sysidList in self.vehicleListing] and compid in [compidList[1] for compidList in self.vehicleListing]):
            self.vehicleListing.append((sysid, compid, 0, m.type))
            # figure out leader for vehicle - check FOLL_SYSID
            self.vehParamsToGet.append((sysid, compid))
            self.needGUIupdate = True
            self.vehicleLastHB[(sysid, compid)] = time.time()
        # Only send these packets on if the vehicle is already in the list
        elif (sysid, compid) in self.vehicleLastHB.keys():
            # update time vehicle was last seen
            if mtype == 'HEARTBEAT':
                self.vehicleLastHB[(sysid, compid)] = time.time()

            # updated leader information from vehicle
            elif mtype == 'PARAM_VALUE' and m.param_id == "FOLL_SYSID":
                for i in range(0, len(self.vehicleListing)):
                    # only update if the leader ID has changed (avoids unessariliy refreshing the UI)
                    if self.vehicleListing[i][0] == sysid and self.vehicleListing[i][1] == compid and self.vehicleListing[i][2] != int(m.param_value):
                        self.vehicleListing[i] = (self.vehicleListing[i][0], self.vehicleListing[i][1], int(
                            m.param_value), self.vehicleListing[i][3])
                        # get GUI to update layout
                        self.needGUIupdate = True
                        break

            # pass to gui elements. Only send relevant packets, otherwise will slow the GUI
            if mtype in ['HEARTBEAT', 'VFR_HUD', 'GLOBAL_POSITION_INT', "SYS_STATUS", "STATUSTEXT", 'PARAM_VALUE']:
                self.gui.onmavlinkpacket(m)


def init(mpstate):
    '''initialise module'''
    return swarm(mpstate)



#!/usr/bin/env python

"""

  MAVProxy realtime graphing module, partly based on the wx graphing
  demo by Eli Bendersky (eliben@gmail.com)

  http://eli.thegreenplace.net/files/prog_code/wx_mpl_dynamic_graph.py.txt
"""

import platform
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc

class LiveGraph():
    '''
    a live graph object using wx and matplotlib
    All of the GUI work is done in a child process to provide some insulation
    from the parent mavproxy instance and prevent instability in the GCS

    New data is sent to the LiveGraph instance via a pipe
    '''
    def __init__(self,
                 fields,
                 title='MAVProxy: LiveGraph',
                 timespan=20.0,
                 tickresolution=0.2,
                 colors=[ 'red', 'green', 'blue', 'orange', 'olive', 'cyan', 'magenta', 'brown',
                          'violet', 'purple', 'grey', 'black']):
        self.fields = fields
        self.colors = colors
        self.title  = title
        self.timespan = timespan
        self.tickresolution = tickresolution
        self.values = [None]*len(self.fields)
        self.parent_pipe,self.child_pipe = multiproc.Pipe()
        self.close_graph = multiproc.Event()
        self.close_graph.clear()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        mp_util.child_close_fds()

        import matplotlib, platform
        if platform.system() != "Darwin":
            # on MacOS we can't set WxAgg here as it conflicts with the MacOS version
            matplotlib.use('WXAgg')

        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx

        app = wx.App(False)
        from MAVProxy.modules.lib import live_graph_ui
        app.frame = live_graph_ui.GraphFrame(state=self)
        app.frame.Show()
        app.MainLoop()
        
    def add_values(self, values):
        '''add some data to the graph'''
        if self.child.is_alive():
            self.parent_pipe.send(values)

    def close(self):
        '''close the graph'''
        self.close_graph.set()
        if self.is_alive():
            self.child.join(2)

    def is_alive(self):
        '''check if graph is still going'''
        return self.child.is_alive()


if __name__ == "__main__":
    multiproc.freeze_support()
    # test the graph
    import time, math
    import live_graph
    livegraph = live_graph.LiveGraph(['sin(t)', 'cos(t)', 'sin(t+1)',
                           'cos(t+1)', 'sin(t+2)', 'cos(t+2)',
                           'cos(t+1)', 'sin(t+2)', 'cos(t+2)', 'x'],
                          timespan=30,
                          title='Graph Test')
    while livegraph.is_alive():
        t = time.time()
        livegraph.add_values([math.sin(t), math.cos(t),
                              math.sin(t+1), math.cos(t+1),
                              math.sin(t+1), math.cos(t+1),
                              math.sin(t+1), math.cos(t+1),
                              math.sin(t+2), math.cos(t+2)])
        time.sleep(0.05)



#!/usr/bin/env python
'''command long'''

import time, os
from numpy import equal
from pymavlink import mavutil
from math import *

from MAVProxy.modules.lib import mp_module

class CmdlongModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(CmdlongModule, self).__init__(mpstate, "cmdlong", public=True)
        self.add_command('setspeed', self.cmd_do_change_speed, "do_change_speed")
        self.add_command('setyaw', self.cmd_condition_yaw, "condition_yaw")
        self.add_command('takeoff', self.cmd_takeoff, "takeoff")
        self.add_command('velocity', self.cmd_velocity, "velocity")
        self.add_command('position', self.cmd_position, "position")
        self.add_command('attitude', self.cmd_attitude, "attitude")
        self.add_command('cammsg', self.cmd_cammsg, "cammsg")
        self.add_command('cammsg_old', self.cmd_cammsg_old, "cammsg_old")
        self.add_command('camctrlmsg', self.cmd_camctrlmsg, "camctrlmsg")
        self.add_command('posvel', self.cmd_posvel, "posvel")
        self.add_command('parachute', self.cmd_parachute, "parachute",
                         ['<enable|disable|release>'])
        self.add_command('long', self.cmd_long, "execute mavlink long command",
                         self.cmd_long_commands())
        self.add_command('command_int', self.cmd_command_int, "execute mavlink command_int",
                         self.cmd_long_commands())
        self.add_command('engine', self.cmd_engine, "engine")
        self.add_command('pause', self.cmd_pause, "pause AUTO/GUIDED modes")
        self.add_command('resume', self.cmd_resume, "resume AUTO/GUIDED modes")

    def cmd_long_commands(self):
        atts = dir(mavutil.mavlink)
        atts = filter( lambda x : x.lower().startswith("mav_cmd"), atts)
        ret = []
        for att in atts:
            ret.append(att)
            ret.append(str(att[8:]))
        return ret

    def cmd_takeoff(self, args):
        '''take off'''
        if ( len(args) != 1):
            print("Usage: takeoff ALTITUDE_IN_METERS")
            return

        if (len(args) == 1):
            altitude = float(args[0])
            print("Take Off started")
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                self.settings.target_component, # target_component
                mavutil.mavlink.MAV_CMD_NAV_TAKEOFF, # command
                0, # confirmation
                0, # param1
                0, # param2
                0, # param3
                0, # param4
                0, # param5
                0, # param6
                altitude) # param7

    def cmd_parachute(self, args):
        '''parachute control'''
        usage = "Usage: parachute <enable|disable|release>"
        if len(args) != 1:
            print(usage)
            return

        cmds = {
            'enable'  : mavutil.mavlink.PARACHUTE_ENABLE,
            'disable' : mavutil.mavlink.PARACHUTE_DISABLE,
            'release' : mavutil.mavlink.PARACHUTE_RELEASE
            }
        if not args[0] in cmds:
            print(usage)
            return
        cmd = cmds[args[0]]
        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            0, # target_component
            mavutil.mavlink.MAV_CMD_DO_PARACHUTE,
            0,
            cmd,
            0, 0, 0, 0, 0, 0)

    def cmd_camctrlmsg(self, args):
        '''camctrlmsg'''

        print("Sent DIGICAM_CONFIGURE CMD_LONG")
        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            0, # target_component
            mavutil.mavlink.MAV_CMD_DO_DIGICAM_CONFIGURE, # command
            0, # confirmation
            10, # param1
            20, # param2
            30, # param3
            40, # param4
            50, # param5
            60, # param6
            70) # param7

    def cmd_cammsg(self, args):
        '''cammsg'''

        params = [0, 0, 0, 0, 1, 0, 0]

        # fill in any args passed by user
        for i in range(min(len(args),len(params))):
            params[i] = float(args[i])

        print("Sent DIGICAM_CONTROL CMD_LONG")
        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            0, # target_component
            mavutil.mavlink.MAV_CMD_DO_DIGICAM_CONTROL, # command
            0, # confirmation
            params[0], # param1
            params[1], # param2
            params[2], # param3
            params[3], # param4
            params[4], # param5
            params[5], # param6
            params[6]) # param7

    def cmd_engine(self, args):
        '''engine control'''
        if len(args) < 1:
            print("usage: engine <1|0>")
            return
        params = [0, 0, 0, 0, 0, 0, 0]

        if args[0] == 'start':
            args[0] = '1'
        if args[0] == 'stop':
            args[0] = '0'
            
        # fill in any args passed by user
        for i in range(min(len(args),len(params))):
            params[i] = float(args[i])

        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            0, # target_component
            mavutil.mavlink.MAV_CMD_DO_ENGINE_CONTROL, # command
            0, # confirmation
            params[0], # param1
            params[1], # param2
            params[2], # param3
            params[3], # param4
            params[4], # param5
            params[5], # param6
            params[6]) # param7

    def cmd_cammsg_old(self, args):
        '''cammsg_old'''

        print("Sent old DIGICAM_CONTROL")
        self.master.mav.digicam_control_send(
            self.settings.target_system,  # target_system
            0, # target_component
            0, 0, 0, 0, 1, 0, 0, 0)

    def cmd_do_change_speed(self, args):
        '''speed value'''
        if ( len(args) != 1):
            print("Usage: setspeed SPEED_VALUE")
            return

        if (len(args) == 1):
            speed = float(args[0])
            print("SPEED %s" % (str(speed)))
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                self.settings.target_component, # target_component
                mavutil.mavlink.MAV_CMD_DO_CHANGE_SPEED, # command
                0, # confirmation
                0, # param1
                speed, # param2 (Speed value)
                0, # param3
                0, # param4
                0, # param5
                0, # param6
                0) # param7

    def cmd_condition_yaw(self, args):
        '''yaw angle angular_speed angle_mode'''
        if ( len(args) != 3):
            print("Usage: yaw ANGLE ANGULAR_SPEED MODE:[0 absolute / 1 relative]")
            return

        if (len(args) == 3):
            angle = float(args[0])
            angular_speed = float(args[1])
            angle_mode = float(args[2])
            print("ANGLE %s" % (str(angle)))
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                self.settings.target_component, # target_component
                mavutil.mavlink.MAV_CMD_CONDITION_YAW, # command
                0, # confirmation
                angle, # param1 (angle value)
                angular_speed, # param2 (angular speed value)
                0, # param3
                angle_mode, # param4 (mode: 0->absolute / 1->relative)
                0, # param5
                0, # param6
                0) # param7

    def cmd_velocity(self, args):
        '''velocity x-ms y-ms z-ms'''
        if (len(args) != 3):
            print("Usage: velocity x y z (m/s)")
            return

        if (len(args) == 3):
            x_mps = float(args[0])
            y_mps = float(args[1])
            z_mps = float(args[2])
            print("x:%f, y:%f, z:%f" % (x_mps, y_mps, z_mps))
            self.master.mav.set_position_target_local_ned_send(
                                      0,  # system time in milliseconds
                                      self.settings.target_system,  # target system
                                      0,  # target component
                                      8,  # coordinate frame MAV_FRAME_BODY_NED
                                      4039,     # type mask (vel only)
                                      0, 0, 0,  # position x,y,z
                                      x_mps, y_mps, z_mps,  # velocity x,y,z
                                      0, 0, 0,  # accel x,y,z
                                      0, 0)     # yaw, yaw rate

    def cmd_position(self, args):
        '''position x-m y-m z-m'''
        if (len(args) != 3):
            print("Usage: position x y z (meters)")
            return

        if (len(args) == 3):
            x_m = float(args[0])
            y_m = float(args[1])
            z_m = float(args[2])
            print("x:%f, y:%f, z:%f" % (x_m, y_m, z_m))
            self.master.mav.set_position_target_local_ned_send(
                                      0,  # system time in milliseconds
                                      self.settings.target_system,  # target system
                                      0,  # target component
                                      8,  # coordinate frame MAV_FRAME_BODY_NED
                                      3576,     # type mask (pos only)
                                      x_m, y_m, z_m,  # position x,y,z
                                      0, 0, 0,  # velocity x,y,z
                                      0, 0, 0,  # accel x,y,z
                                      0, 0)     # yaw, yaw rate

    def cmd_attitude(self, args):
        '''attitude mask q0 q1 q2 q3 roll_rate pitch_rate yaw_rate thrust'''
        if len(args) < 5:
            print("Usage: attitude q0 q1 q2 q3 thrust")
            print("q0 q1 q2 q3: [w, x, y, z] order, zero-rotation is [1, 0, 0, 0], unit-length")
            print("thrust: (0~1)")
            return
        elif len(args) not in [5, 9]:
            print("Usage: attitude mask q0 q1 q2 q3 roll_rate pitch_rate yaw_rate thrust")
            print("mask : Example 7 (0b00000111): Ignore roll rate, pitch rate, yaw rate")
            print("mask : Example 128 (0b10000000): Ignore attitude")
            print("mask : Example 132 (0b10000100): Ignore yaw rate, Ignore attitude")
            print("     : See https://mavlink.io/en/messages/common.html#ATTITUDE_TARGET_TYPEMASK")
            print("q0 q1 q2 q3: [w, x, y, z] order, zero-rotation is [1, 0, 0, 0], unit-length")
            print("roll_rate pitch_rate yaw_rate: in degrees per second")
            print("thrust: (0~1)")
            return

        if len(args) == 5:
            mask = 7                    # ignore angular rates
            q0 = float(args[0])
            q1 = float(args[1])
            q2 = float(args[2])
            q3 = float(args[3])
            thrust = float(args[4])
            roll_rate = 0.0
            pitch_rate = 0.0
            yaw_rate = 0.0
            print("q0:%.3f, q1:%.3f, q2:%.3f q3:%.3f, thrust:%.3f" % (q0, q1, q2, q3, thrust))

        elif len(args) == 9:
            mask = int(args[0])
            q0 = float(args[1])
            q1 = float(args[2])
            q2 = float(args[3])
            q3 = float(args[4])
            roll_rate = float(args[5])
            pitch_rate = float(args[6])
            yaw_rate = float(args[7])
            thrust = float(args[8])
            print("mask:%i, q0:%.3f, q1:%.3f, q2:%.3f q3:%.3f, roll_rate:%.3f, pitch_rate:%.3f, yaw_rate:%.3f, thrust:%.3f" %
                  (mask, q0, q1, q2, q3, roll_rate, pitch_rate, yaw_rate, thrust))

        att_target = [q0, q1, q2, q3]
        self.master.mav.set_attitude_target_send(
                                    0,  # system time in milliseconds
                                    self.settings.target_system,  # target system
                                    0,            # target component
                                    mask,         # type mask
                                    att_target,   # quaternion attitude
                                    radians(roll_rate),    # body roll rate
                                    radians(pitch_rate),   # body pitch rate
                                    radians(yaw_rate),     # body yaw rate
                                    thrust)       # thrust

    def cmd_posvel(self, args):
        '''posvel mapclick vN vE vD'''
        ignoremask = 511
        latlon = None
        latlon = self.mpstate.click_location
        if latlon is None:
            print("set latlon to zeros")
            latlon = [0, 0]
        else:
            ignoremask = ignoremask & 504
            print("found latlon", ignoremask)
        vN = 0
        vE = 0
        vD = 0
        if (len(args) == 3):
            vN = float(args[0])
            vE = float(args[1])
            vD = float(args[2])
            ignoremask = ignoremask & 455

        print("ignoremask",ignoremask)
        print(latlon)
        self.master.mav.set_position_target_global_int_send(
            0,  # system time in ms
            self.settings.target_system,  # target system
            0,  # target component
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT_INT,
            ignoremask, # ignore
            int(latlon[0] * 1e7),
            int(latlon[1] * 1e7),
            10,
            vN, vE, vD, # velocity
            0, 0, 0, # accel x,y,z
            0, 0) # yaw, yaw rate

    def cmd_pause(self, args):
        '''pause AUTO/GUIDED modes'''
        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            self.settings.target_component,  # target_component
            mavutil.mavlink.MAV_CMD_DO_PAUSE_CONTINUE,  # command
            0,  # confirmation
            0,  # 0: pause, 1: continue
            0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0)  # param7

    def cmd_resume(self, args):
        '''resume AUTO/GUIDED modes'''
        self.master.mav.command_long_send(
            self.settings.target_system,  # target_system
            self.settings.target_component,  # target_component
            mavutil.mavlink.MAV_CMD_DO_PAUSE_CONTINUE,  # command
            0,  # confirmation
            1,  # 0: pause, 1: continue
            0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0)  # param7

    def cmd_long(self, args):
        '''execute supplied command long'''
        if len(args) < 1:
            print("Usage: long <command> [arg1] [arg2]...")
            return
        command = None
        if args[0].isdigit():
            command = int(args[0])
        else:
            try:
                command = getattr(mavutil.mavlink, args[0])
            except AttributeError as e:
                try:
                    command = getattr(mavutil.mavlink, "MAV_CMD_" + args[0])
                except AttributeError as e:
                    pass

        if command is None:
            print("Unknown command long ({0})".format(args[0]))
            return

        floating_args = [ float(x) for x in args[1:] ]
        while len(floating_args) < 7:
            floating_args.append(float(0))
        self.master.mav.command_long_send(self.settings.target_system,
                                          self.settings.target_component,
                                          command,
                                          0,
                                          *floating_args)

    def cmd_command_int(self, args):
        '''execute supplied command_int'''
        want_args = 11
        if len(args) != want_args:
            print("Argument count issue: want={0} got={1}".format(want_args, len(args)))
            print("Usage: command_int frame command current autocontinue param1 param2 param3 param4 x y z")
            print("e.g. command_int GLOBAL_RELATIVE_ALT DO_SET_HOME 0 0 0 0 0 0 -353632120 1491659330 0")
            print("e.g. command_int GLOBAL MAV_CMD_DO_SET_ROI 0 0 0 0 0 0 5000000 5000000 500")
            return

        frame = None
        if args[0].isdigit():
            frame = int(args[0])
        else:
            try:
                # attempt to allow MAV_FRAME_GLOBAL for frame
                frame = getattr(mavutil.mavlink, args[0])
            except AttributeError as e:
                try:
                    # attempt to allow GLOBAL for frame
                    frame = getattr(mavutil.mavlink, "MAV_FRAME_" + args[0])
                except AttributeError as e:
                    pass

        if frame is None:
            print("Unknown frame ({0})".format(args[0]))
            return

        command = None
        if args[1].isdigit():
            command = int(args[1])
        else:
            # let "command_int ... MAV_CMD_DO_SET_HOME ..." work
            try:
                command = getattr(mavutil.mavlink, args[1])
            except AttributeError as e:
                try:
                    # let "command_int ... DO_SET_HOME" work
                    command = getattr(mavutil.mavlink, "MAV_CMD_" + args[1])
                except AttributeError as e:
                    pass

        current = int(args[2])
        autocontinue = int(args[3])
        param1 = float(args[4])
        param2 = float(args[5])
        param3 = float(args[6])
        param4 = float(args[7])
        x = int(args[8])
        y = int(args[9])
        z = float(args[10])
        self.master.mav.command_int_send(self.settings.target_system,
                                         self.settings.target_component,
                                         frame,
                                         command,
                                         0,
                                         0,
                                         param1,
                                         param2,
                                         param3,
                                         param4,
                                         x,
                                         y,
                                         z)

def init(mpstate):
    '''initialise module'''
    return CmdlongModule(mpstate)



"""
  MAVProxy console

  uses lib/console.py for display
"""

from MAVProxy.modules.lib import wxhorizon
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib.wxhorizon_util import Attitude, VFR_HUD, Global_Position_INT, BatteryInfo, FlightState, WaypointInfo, FPS

import time

class HorizonModule(mp_module.MPModule):
    def __init__(self, mpstate):
        # Define module load/unload reference and window title
        super(HorizonModule, self).__init__(mpstate, "horizon", "Horizon Indicator", public=True)
        self.mpstate.horizonIndicator = wxhorizon.HorizonIndicator(title='Horizon Indicator')
        self.mode = ''
        self.armed = ''
        self.currentWP = 0
        self.finalWP = 0
        self.currentDist = 0
        self.nextWPTime = 0
        self.speed = 0
        self.wpBearing = 0
        self.msgList = []
        self.lastSend = 0.0
        self.fps = 10.0
        self.sendDelay = (1.0/self.fps)*0.9
        self.add_command('horizon-fps',self.fpsInformation,"Get or change frame rate for horizon. Usage: horizon-fps set <fps>, horizon-fps get. Set fps to zero to get unrestricted framerate.")
        
    def unload(self):
        '''unload module'''
        self.mpstate.horizonIndicator.close()
            
    def fpsInformation(self,args):
        '''fps command'''
        invalidStr = 'Invalid number of arguments. Usage horizon-fps set <fps> or horizon-fps get. Set fps to zero to get unrestricted framerate.'
        if len(args)>0:
            if args[0] == "get":
                '''Get the current framerate.'''
                if (self.fps == 0.0):
                    print('Horizon Framerate: Unrestricted')
                else:
                    print("Horizon Framerate: " + str(self.fps))
            elif args[0] == "set":
                if len(args)==2:
                    self.fps = float(args[1])
                    if (self.fps != 0):
                        self.sendDelay = 1.0/self.fps
                    else:
                        self.sendDelay = 0.0
                    self.msgList.append(FPS(self.fps))
                    if (self.fps == 0.0):
                        print('Horizon Framerate: Unrestricted')
                    else:
                        print("Horizon Framerate: " + str(self.fps))
                else:
                    print(invalidStr)
            else:
                print(invalidStr)
        else:
            print(invalidStr)
            
    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        msgType = msg.get_type()
        master = self.master       
        if msgType == 'HEARTBEAT':
            # Update state and mode information
            self.armed = master.motors_armed()
            self.mode = master.flightmode
            # Send Flight State information down pipe
            self.msgList.append(FlightState(self.mode,self.armed))
        elif msgType == 'ATTITUDE':
            # Send attitude information down pipe
            self.msgList.append(Attitude(msg))
        elif msgType == 'VFR_HUD':
            # Send HUD information down pipe
            self.msgList.append(VFR_HUD(msg))
        elif msgType == 'GLOBAL_POSITION_INT':
            # Send altitude information down pipe
            self.msgList.append(Global_Position_INT(msg,time.time()))
        elif msgType == 'SYS_STATUS':
            # Mode and Arm State
            self.msgList.append(BatteryInfo(msg))
        elif msgType in ['WAYPOINT_CURRENT', 'MISSION_CURRENT']:
            # Waypoints
            self.currentWP = msg.seq
            self.finalWP = self.module('wp').wploader.count()
            self.msgList.append(WaypointInfo(self.currentWP,self.finalWP,self.currentDist,self.nextWPTime,self.wpBearing))
        elif msgType == 'NAV_CONTROLLER_OUTPUT':
            self.currentDist = msg.wp_dist
            self.speed = master.field('VFR_HUD', 'airspeed', 30)
            if self.speed > 1:
                self.nextWPTime = self.currentDist / self.speed
            else:
                self.nextWPTime = '-'
            self.wpBearing = msg.target_bearing
            self.msgList.append(WaypointInfo(self.currentWP,self.finalWP,self.currentDist,self.nextWPTime,self.wpBearing))

    
    def idle_task(self):
        if self.mpstate.horizonIndicator.close_event.wait(0.001):
            self.needs_unloading = True   # tell MAVProxy to unload this module
    
        if (time.time() - self.lastSend) > self.sendDelay:
            self.mpstate.horizonIndicator.parent_pipe_send.send(self.msgList)
            self.msgList = []
            self.lastSend = time.time()
    
def init(mpstate):
    '''initialise module'''
    return HorizonModule(mpstate)



#!/usr/bin/env python
'''
gimbal control module
Andrew Tridgell
January 2015
'''

import sys, os, time
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.mavproxy_map import mp_slipmap
from pymavlink import mavutil
from pymavlink.rotmat import Vector3, Matrix3, Plane, Line
from math import radians
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

class GimbalModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(GimbalModule, self).__init__(mpstate, "gimbal", "gimbal control module")
        self.add_command('gimbal', self.cmd_gimbal, "gimbal link control",
                         ['<rate|point|roi|roivel|mode|status>'])
        if mp_util.has_wxpython:
            self.menu = MPMenuSubMenu('Mount',
                                      items=[MPMenuItem('GPS targeting Mode', 'GPS Mode', '# gimbal mode gps'),
                                             MPMenuItem('MAVLink targeting Mode', 'MAVLink Mode', '# gimbal mode mavlink'),
                                             MPMenuItem('RC targeting Mode', 'RC Mode', '# gimbal mode rc'),
                                             MPMenuItem('Point At', 'Point At', '# gimbal roi')])
            self.menu_added_map = False
        else:
            self.menu = None

    def idle_task(self):
        '''called on idle'''
        if self.menu is not None and self.module('map') is not None and not self.menu_added_map:
            self.menu_added_map = True
            self.module('map').add_menu(self.menu)

    def cmd_gimbal(self, args):
        '''control gimbal'''
        usage = 'Usage: gimbal <rate|point|roi|roivel|mode|status>'
        if len(args) == 0:
            print(usage)
            return
        if args[0] == 'rate':
            self.cmd_gimbal_rate(args[1:])
        elif args[0] == 'point':
            self.cmd_gimbal_point(args[1:])
        elif args[0] == 'roi':
            self.cmd_gimbal_roi(args[1:])
        elif args[0] == 'mode':
            self.cmd_gimbal_mode(args[1:])
        elif args[0] == 'status':
            self.cmd_gimbal_status(args[1:])
        elif args[0] == 'roivel':
            self.cmd_gimbal_roi_vel(args[1:])

    def cmd_gimbal_mode(self, args):
        '''control gimbal mode'''
        if len(args) != 1:
            print("usage: gimbal mode <RETRACT|NEUTRAL|GPS|MAVLink|RC>")
            return
        if args[0].upper() == "RETRACT":
            mode = mavutil.mavlink.MAV_MOUNT_MODE_RETRACT
        elif args[0].upper() == "NEUTRAL":
            mode = mavutil.mavlink.MAV_MOUNT_MODE_NEUTRAL
        elif args[0].upper() == 'GPS':
            mode = mavutil.mavlink.MAV_MOUNT_MODE_GPS_POINT
        elif args[0].upper() == 'MAVLINK':
            mode = mavutil.mavlink.MAV_MOUNT_MODE_MAVLINK_TARGETING
        elif args[0].upper() == 'RC':
            mode = mavutil.mavlink.MAV_MOUNT_MODE_RC_TARGETING
        else:
            print("Unsupported mode %s" % args[0])
        self.master.mav.mount_configure_send(self.target_system,
                                             self.target_component,
                                             mode,
                                             1, 1, 1)

    def cmd_gimbal_roi(self, args):
        '''control roi position'''
        latlon = None
        alt = 0
        if len(args) == 0:
            latlon = self.mpstate.click_location
        elif len(args) == 3:
            latlon = [float(i) for i in args[0:2]]
            alt = float(args[2])
        if latlon is None and len(args) == 0:
            print("No map click position available and no parameters set")
            return
        elif latlon is None:
            print("usage: gimbal roi [LAT LON ALT]")
            return
        self.master.mav.mount_control_send(self.target_system,
                                           self.target_component,
                                           int(latlon[0]*1e7),
                                           int(latlon[1]*1e7),
                                           int(alt),
                                           0)

    def cmd_gimbal_roi_vel(self, args):
        '''control roi position and velocity'''
        if len(args) != 0 and len(args) != 3 and len(args) != 6:
            print("usage: gimbal roivel [VEL_NORTH VEL_EAST VEL_DOWN] [ACC_NORTH ACC_EASY ACC_DOWN]")
            return
        latlon = None
        vel = [0,0,0]
        acc = [0,0,0]
        if (len(args) >= 3):
            vel[0:3] = args[0:3]
        if (len(args) == 6):
            acc[0:3] = args[3:6]
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            latlon = (0,0,0)
        self.master.mav.set_roi_global_int_send(0, #time_boot_ms
            1, #target_system
            1, #target_component
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT_INT,
            0, #type_mask
            0, #roi_index
            0, #timeout_ms
            int(latlon[0]*1e7), #lat int
            int(latlon[1]*1e7), #lng int
            float(0),       #alt
            float(vel[0]), #vx
            float(vel[1]), #vy
            float(vel[2]), #vz
            float(acc[0]), #ax
            float(acc[1]), #ay
            float(acc[2])) #az

    def cmd_gimbal_rate(self, args):
        '''control gimbal rate'''
        if len(args) != 3:
            print("usage: gimbal rate ROLL PITCH YAW")
            return
        (roll, pitch, yaw) = (float(args[0]), float(args[1]), float(args[2]))
        self.master.mav.gimbal_control_send(self.target_system,
                                            mavutil.mavlink.MAV_COMP_ID_GIMBAL,
                                            radians(roll),
                                            radians(pitch),
                                            radians(yaw))

    def cmd_gimbal_point(self, args):
        '''control gimbal pointing'''
        if len(args) != 3:
            print("usage: gimbal point ROLL PITCH YAW")
            return
        (roll, pitch, yaw) = (float(args[0]), float(args[1]), float(args[2]))
        self.master.mav.mount_control_send(self.target_system,
                                           self.target_component,
                                           int(pitch*100),
                                           int(roll*100),
                                           int(yaw*100),
                                           0)

    def cmd_gimbal_status(self, args):
        '''show gimbal status'''
        master = self.master
        if 'GIMBAL_REPORT' in master.messages:
            print(master.messages['GIMBAL_REPORT'])
        else:
            print("No GIMBAL_REPORT messages")

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''

        if not self.mpstate.map:
            # don't draw if no map
            return

        if m.get_type() != 'GIMBAL_REPORT':
            return

        needed = ['ATTITUDE', 'GLOBAL_POSITION_INT']
        for n in needed:
            if not n in self.master.messages:
                return

        # clear the camera icon
        self.mpstate.map.add_object(mp_slipmap.SlipClearLayer('GimbalView'))

        gpi = self.master.messages['GLOBAL_POSITION_INT']
        att = self.master.messages['ATTITUDE']
        vehicle_dcm = Matrix3()
        vehicle_dcm.from_euler(att.roll, att.pitch, att.yaw)

        rotmat_copter_gimbal = Matrix3()
        rotmat_copter_gimbal.from_euler312(m.joint_roll, m.joint_el, m.joint_az)
        gimbal_dcm = vehicle_dcm * rotmat_copter_gimbal

        lat = gpi.lat * 1.0e-7
        lon = gpi.lon * 1.0e-7
        alt = gpi.relative_alt * 1.0e-3

        # ground plane
        ground_plane = Plane()

        # the position of the camera in the air, remembering its a right
        # hand coordinate system, so +ve z is down
        camera_point = Vector3(0, 0, -alt)

        # get view point of camera when not rotated
        view_point = Vector3(1, 0, 0)

        # rotate view_point to form current view vector
        rot_point = gimbal_dcm * view_point

        # a line from the camera to the ground
        line = Line(camera_point, rot_point)

        # find the intersection with the ground
        pt = line.plane_intersection(ground_plane, forward_only=True)
        if pt is None:
            # its pointing up into the sky
            return None

        (view_lat, view_lon) = mp_util.gps_offset(lat, lon, pt.y, pt.x)

        icon = self.mpstate.map.icon('camera-small-red.png')
        self.mpstate.map.add_object(mp_slipmap.SlipIcon('gimbalview',
                                                        (view_lat,view_lon),
                                                        icon, layer='GimbalView', rotation=0, follow=False))


def init(mpstate):
    '''initialise module'''
    return GimbalModule(mpstate)



"""
control EMU ECU system
"""

import time

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

class EMUECUModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(EMUECUModule, self).__init__(mpstate, "emuecu", "emuecu", public=False)
        self.emuecu_settings = mp_settings.MPSettings(
            [('port', int, 102)])
        self.add_command('emu', self.cmd_emu, 'EMUECU control',
                         ["<send>",
                          "set (EMUECUSETTING)"])
        self.add_completion_function('(EMUECUSETTING)',
                                     self.emuecu_settings.completion)

    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        if msg.get_type() == 'SERIAL_CONTROL':
            print(msg)

    def cmd_emu(self, args):
        '''emu command handling'''
        if len(args) <= 0:
            print("Usage: emu <send|set>")
            return
        if args[0] == "send":
            self.cmd_send(args[1:])
        elif args[0] == "set":
            self.emuecu_settings.command(args[1:])

    def cmd_send(self, args):
        '''send command'''
        cmd = ' '.join(args) + '\n'
        buf = [ord(x) for x in cmd]
        buf.extend([0]*(70-len(buf)))
        mav = self.master.mav
        mav.serial_control_send(self.emuecu_settings.port,
                                0,
                                0, 0,
                                len(cmd), buf)


def init(mpstate):
    '''initialise module'''
    return EMUECUModule(mpstate)



import BaseHTTPServer
import json
import os.path
import thread
import urlparse

DOC_DIR = os.path.join(os.path.dirname(__file__), 'mmap_app')


class Server(BaseHTTPServer.HTTPServer):
  def __init__(self, handler, address='', port=9999, module_state=None):
    BaseHTTPServer.HTTPServer.__init__(self, (address, port), handler)
    self.allow_reuse_address = True
    self.module_state = module_state


class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
  def do_GET(self):
    scheme, host, path, params, query, frag = urlparse.urlparse(self.path)
    if path == '/data':
      state = self.server.module_state
      data = {'lat': state.lat,
              'lon': state.lon,
              'heading': state.heading,
              'alt': state.alt,
              'airspeed': state.airspeed,
              'groundspeed': state.groundspeed}
      self.send_response(200)
      self.end_headers()
      self.wfile.write(json.dumps(data))
    else:
      # Remove leading '/'.
      path = path[1:]
      # Ignore all directories.  E.g.  for ../../bar/a.txt serve
      # DOC_DIR/a.txt.
      unused_head, path = os.path.split(path)
      # for / serve index.html.
      if path == '':
        path = 'index.html'
      content = None
      error = None
      try:
        import pkg_resources
        name = __name__
        if name == "__main__":
          name = "MAVProxy.modules.mavproxy_mmap.????"
        content = pkg_resources.resource_stream(name, "mmap_app/%s" % path).read()
      except IOError as e:
        error = str(e)
      if content:
        self.send_response(200)
        self.end_headers()
        self.wfile.write(content)
      else:
        self.send_response(404)
        self.end_headers()
        self.wfile.write('Error: %s' % (error,))


def start_server(address, port, module_state):
  server = Server(
    Handler, address=address, port=port, module_state=module_state)
  thread.start_new_thread(server.serve_forever, ())
  return server
